# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2014-2015  Jon Lund Steffensen <jon_steffensen@uri.edu>

"""Representations of affine expressions and variables.

These classes can be used to represent affine expressions
and do manipulation and evaluation with substitutions of
particular variables.
"""

import re
import numbers
import functools
from collections import Counter, defaultdict

from six import string_types, iteritems, iterkeys, itervalues


@functools.total_ordering
class Variable(object):
    """Represents a variable in an expression

    Equality of variables is based on the symbol.
    """

    def __init__(self, symbol):
        """Create variable with given symbol

        Symbol must start with a letter or underscore but
        can contain numbers in other positions.

        >>> Variable('x')
        Variable('x')
        """

        if not re.match(r'^[^\d\W]\w*\Z', symbol):
            raise ValueError('Invalid symbol {}'.format(symbol))
        self._symbol = str(symbol)

    @property
    def symbol(self):
        """Symbol of variable

        >>> Variable('x').symbol
        'x'
        """
        return self._symbol

    def simplify(self):
        """Return simplified expression

        The simplified form of a variable is always the
        variable itself.

        >>> Variable('x').simplify()
        Variable('x')
        """
        return self

    def substitute(self, mapping):
        """Return expression with variables substituted

        >>> Variable('x').substitute(lambda v: {'x': 567}.get(v.symbol, v))
        567
        >>> Variable('x').substitute(lambda v: {'y': 42}.get(v.symbol, v))
        Variable('x')
        >>> Variable('x').substitute(
        ...     lambda v: {'x': 123, 'y': 56}.get(v.symbol, v))
        123
        """

        return mapping(self)

    def __add__(self, other):
        return Expression({self: 1}) + other

    def __radd__(self, other):
        return self + other

    def __sub__(self, other):
        return Expression({self: 1}) - other

    def __rsub__(self, other):
        return -self + other

    def __mul__(self, other):
        return Expression({self: 1}) * other

    def __rmul__(self, other):
        return self * other

    def __div__(self, other):
        return Expression({self: 1}) / other

    __truediv__ = __div__

    def __floordiv__(self, other):
        return Expression({self: 1}) // other

    def __neg__(self):
        return Expression({self: -1})

    def __eq__(self, other):
        """Check equality of variables"""
        if isinstance(other, Expression):
            return other == self
        return isinstance(other, Variable) and self._symbol == other._symbol

    def __ne__(self, other):
        return not self == other

    def __lt__(self, other):
        if isinstance(other, Variable):
            return self._symbol < other._symbol
        return NotImplemented

    def __hash__(self):
        return hash('Variable') ^ hash(self._symbol)

    def __str__(self):
        return self._symbol

    def __repr__(self):
        return 'Variable({})'.format(repr(self._symbol))


class Expression(object):
    """Represents an affine expression (e.g. 2x + 3y - z + 5)"""

    def __init__(self, *args):
        """Create new expression

        >>> Expression({ Variable('x'): 2 }, 3)
        Expression('2x + 3')
        >>> Expression({ Variable('x'): 1, Variable('y'): 1 })
        Expression('x + y')
        """

        if len(args) == 1 and isinstance(args[0], string_types):
            # Parse as string
            self._variables, self._offset = self._parse_string(args[0])
        elif len(args) <= 2:
            self._variables = {}
            self._offset = args[1] if len(args) >= 2 else 0

            variables = args[0] if len(args) >= 1 else {}
            for var, value in iteritems(variables):
                if not isinstance(var, Variable):
                    raise ValueError('Not a variable: {}'.format(var))
                if value != 0:
                    self._variables[var] = value
        else:
            raise TypeError('Expression() requires one or two arguments')

    def _parse_string(self, s):
        """Parse expression string

        Variables must be valid variable symbols and
        coefficients must be integers.
        """

        scanner = re.compile(r'''
            (\s+) |         # whitespace
            ([^\d\W]\w*) |  # variable
            (\d+) |         # number
            ([+-]) |        # sign
            (\Z) |          # end
            (.)             # error
        ''', re.DOTALL | re.VERBOSE)

        variables = defaultdict(int)
        offset = 0

        # Parse using four states:
        # 0: expect sign, variable, number or end (start state)
        # 1: expect sign or end
        # 2: expect variable or number
        # 3: expect sign, variable or end
        # All whitespace is ignored
        state = 0
        state_number = 1
        for match in re.finditer(scanner, s):
            whitespace, variable, number, sign, end, error = match.groups()
            if error is not None:
                raise ValueError(
                    'Invalid token in expression string: {}'.format(
                        match.group(0)))
            elif whitespace is not None:
                continue
            elif variable is not None and state in (0, 2, 3):
                variables[Variable(variable)] += state_number
                state = 1
            elif sign is not None and state in (0, 1, 3):
                if state == 3:
                    offset += state_number
                state_number = 1 if sign == '+' else -1
                state = 2
            elif number is not None and state in (0, 2):
                state_number = state_number * int(number)
                state = 3
            elif end is not None and state in (0, 1, 3):
                if state == 3:
                    offset += state_number
            else:
                raise ValueError(
                    'Invalid token in expression string: {}'.format(
                        match.group(0)))

        # Remove zero-coefficient elements
        variables = {var: value for var, value in iteritems(variables)
                     if value != 0}
        return variables, offset

    def simplify(self):
        """Return simplified expression.

        If the expression is of the form 'x', the variable will be returned,
        and if the expression contains no variables, the offset will be
        returned as a number.
        """
        result = self.__class__(self._variables, self._offset)
        if len(result._variables) == 0:
            return result._offset
        elif len(result._variables) == 1 and result._offset == 0:
            var, value = next(iteritems(result._variables))
            if value == 1:
                return var
        return result

    def substitute(self, mapping):
        """Return expression with variables substituted

        >>> Expression('x + 2y').substitute(
        ...     lambda v: {'y': -3}.get(v.symbol, v))
        Expression('x - 6')
        >>> Expression('x + 2y').substitute(
        ...     lambda v: {'y': Variable('z')}.get(v.symbol, v))
        Expression('x + 2z')
        """
        expr = self.__class__()
        for var, value in iteritems(self._variables):
            expr += value * var.substitute(mapping)
        return (expr + self._offset).simplify()

    def variables(self):
        """Return iterator of variables in expression"""
        return iter(self._variables)

    def __add__(self, other):
        """Add expressions, variables or numbers"""

        if isinstance(other, numbers.Number):
            return self.__class__(self._variables, self._offset + other)
        elif isinstance(other, Variable):
            return self + Expression({other: 1})
        elif isinstance(other, Expression):
            variables = Counter(self._variables)
            variables.update(other._variables)
            variables = {var: value for var, value in iteritems(variables)
                         if value != 0}
            return self.__class__(variables, self._offset + other._offset)
        return NotImplemented

    def __radd__(self, other):
        return self + other

    def __sub__(self, other):
        """Subtract expressions, variables or numbers"""
        return self + -other

    def __rsub__(self, other):
        return -self + other

    def __mul__(self, other):
        """Multiply by scalar"""
        if isinstance(other, numbers.Number):
            if other == 0:
                return self.__class__()
            return self.__class__({var: value * other for var, value
                                   in iteritems(self._variables)},
                                  self._offset * other)
        return NotImplemented

    def __rmul__(self, other):
        return self * other

    def __div__(self, other):
        """Divide by scalar"""
        if isinstance(other, numbers.Number):
            return self.__class__({var: value / other for var, value
                                   in iteritems(self._variables)},
                                  self._offset / other)
        return NotImplemented

    __truediv__ = __div__

    def __floordiv__(self, other):
        if isinstance(other, numbers.Number):
            return self.__class__({var: value // other for var, value
                                   in iteritems(self._variables)},
                                  self._offset // other)
        return NotImplemented

    def __neg__(self):
        return self * -1

    def __eq__(self, other):
        """Expression equality"""
        if isinstance(other, Expression):
            return (self._variables == other._variables and
                    self._offset == other._offset)
        elif isinstance(other, Variable):
            # Check that there is just one variable in the expression
            # with a coefficient of one.
            return (self._offset == 0 and len(self._variables) == 1 and
                    next(iterkeys(self._variables)) == other and
                    next(itervalues(self._variables)) == 1)
        elif isinstance(other, numbers.Number):
            return len(self._variables) == 0 and self._offset == other
        return False

    def __ne__(self, other):
        return not self == other

    def __str__(self):
        def all_terms():
            count_vars = 0
            for symbol, value in sorted((var.symbol, value) for var, value
                                        in iteritems(self._variables)):
                if value != 0:
                    count_vars += 1
                    yield symbol, value
            if self._offset != 0 or count_vars == 0:
                yield None, self._offset

        terms = []
        for i, spec in enumerate(all_terms()):
            symbol, value = spec
            if i == 0:
                # First term is special
                if symbol is None:
                    terms.append('{}'.format(value))
                elif abs(value) == 1:
                    terms.append(symbol if value > 0 else '-'+symbol)
                else:
                    terms.append('{}{}'.format(value, symbol))
            else:
                prefix = '+' if value >= 0 else '-'
                if symbol is None:
                    terms.append('{} {}'.format(prefix, abs(value)))
                elif abs(value) == 1:
                    terms.append('{} {}'.format(prefix, symbol))
                else:
                    terms.append('{} {}{}'.format(prefix, abs(value), symbol))
        return ' '.join(terms)

    def __repr__(self):
        return 'Expression({})'.format(repr(str(self)))


if __name__ == '__main__':
    import doctest
    doctest.testmod()



# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2014-2015  Jon Lund Steffensen <jon_steffensen@uri.edu>

"""Identify blocked metabolites and possible reconstructions.

This implements a variant of the algorithms described in [Kumar07]_.
"""

from six import iteritems

from .lpsolver import lp


class GapFillError(Exception):
    """Indicates an error while running GapFind/GapFill"""


def gapfind(model, solver, epsilon=0.001, v_max=1000):
    """Identify compounds in the model that cannot be produced.

    Yields all compounds that cannot be produced. This method
    assumes implicit sinks for all compounds in the model so
    the only factor that influences whether a compound can be
    produced is the presence of the compounds needed to produce it.

    Epsilon indicates the threshold amount of a compound produced
    for it to not be considered blocked. V_max indicates the
    maximum flux.

    This method is implemented as a MILP-program. Therefore it may
    not be efficient for larger models.
    """
    prob = solver.create_problem()

    # Define flux variables
    for reaction_id in model.reactions:
        lower, upper = model.limits[reaction_id]
        prob.define(('v', reaction_id), lower=lower, upper=upper)

    # Define constraints on production of metabolites in reaction
    binary_cons_lhs = {compound: 0 for compound in model.compounds}
    for spec, value in iteritems(model.matrix):
        compound, reaction_id = spec
        if value != 0 and (reaction_id in model.reversible or value > 0):
            prob.define(('w', reaction_id, compound),
                        types=lp.VariableType.Binary)

            w = prob.var(('w', reaction_id, compound))
            sv = float(value) * prob.var(('v', reaction_id))

            prob.add_linear_constraints(sv <= v_max*w)
            if model.is_reversible(reaction_id):
                prob.add_linear_constraints(sv >= epsilon-v_max*(1 - w))
            else:
                prob.add_linear_constraints(sv >= epsilon*w)

            binary_cons_lhs[compound] += w

    prob.define(*(('xp', compound) for compound in model.compounds),
                types=lp.VariableType.Binary)

    objective = prob.expr(
        {('xp', compound): 1 for compound in model.compounds})
    prob.set_linear_objective(objective)

    for compound, lhs in iteritems(binary_cons_lhs):
        prob.add_linear_constraints(lhs >= prob.var(('xp', compound)))

    # Define mass balance constraints
    massbalance_lhs = {compound: 0 for compound in model.compounds}
    for spec, value in iteritems(model.matrix):
        compound, reaction_id = spec
        massbalance_lhs[compound] += prob.var(('v', reaction_id)) * value
    for compound, lhs in iteritems(massbalance_lhs):
        # The constraint is merely >0 meaning that we have implicit sinks
        # for all compounds.
        prob.add_linear_constraints(lhs >= 0)

    # Solve
    result = prob.solve(lp.ObjectiveSense.Maximize)
    if not result:
        raise GapFillError('Non-optimal solution: {}'.format(result.status))

    for compound in model.compounds:
        if result.get_value(('xp', compound)) == 0:
            yield compound


def gapfill(model, core, blocked, solver, epsilon=0.001, v_max=1000):
    """Find a set of reactions to add such that no compounds are blocked.

    Returns two iterators: first an iterator of reactions not in
    core, that were added to resolve the model. Second, an
    iterator of reaction in core, that were reversed. Similarly to
    GapFind, this method assumes implicit sinks for all compounds in
    the model so the only factor that influences whether a compound
    can be produced is the presence of the compounds needed to produce
    it. This means that the resulting model will not necessarily be
    flux consistent.

    Core indicates the core set of reactions in the model. GapFill will
    minimize the number of added reactions that are not in core. Blocked
    indicates the set of compounds to be resolved. Epsilon indicates the
    threshold amount of a compound produced for it to not be considered
    blocked. V_max indicates the maximum flux.

    This method is implemented as a MILP-program. Therefore it may
    not be efficient for larger models.
    """
    prob = solver.create_problem()

    # Define flux variables
    prob.define(*(('v', reaction_id) for reaction_id in model.reactions),
                lower=-v_max, upper=v_max)

    # Add binary indicator variables
    database_reactions = set(model.reactions).difference(core)
    prob.define(*(('ym', reaction_id) for reaction_id in core),
                types=lp.VariableType.Binary)
    prob.define(*(('yd', reaction_id) for reaction_id in database_reactions),
                types=lp.VariableType.Binary)

    objective = prob.expr(
        {('ym', reaction_id): 1 for reaction_id in core})
    objective += prob.expr(
        {('yd', reaction_id): 1 for reaction_id in database_reactions})
    prob.set_linear_objective(objective)

    # Add constraints on core reactions
    for reaction_id in core:
        v = prob.var(('v', reaction_id))
        if model.is_reversible(reaction_id):
            prob.add_linear_constraints(v >= model.limits[reaction_id].lower)
        else:
            ym = prob.var(('ym', reaction_id))
            prob.add_linear_constraints(v >= -v_max * ym)
        prob.add_linear_constraints(v <= model.limits[reaction_id].upper)

    # Add constraints on database reactions
    for reaction_id in database_reactions:
        v = prob.var(('v', reaction_id))
        yd = prob.var(('yd', reaction_id))
        prob.add_linear_constraints(
            v >= yd * model.limits[reaction_id].lower,
            v <= yd * model.limits[reaction_id].upper)

    # Define constraints on production of blocked metabolites in reaction
    binary_cons_lhs = {compound: 0 for compound in blocked}
    for (compound, reaction_id), value in iteritems(model.matrix):
        if compound in blocked and value != 0:
            prob.define(('w', reaction_id, compound),
                        types=lp.VariableType.Binary)

            w = prob.var(('w', reaction_id, compound))
            sv = float(value) * prob.var(('v', reaction_id))

            prob.add_linear_constraints(sv >= epsilon - v_max * (1 - w))
            prob.add_linear_constraints(sv <= v_max * w)

            if model.is_reversible(reaction_id) or value > 0:
                binary_cons_lhs[compound] += w
            elif reaction_id in core:
                # In this case, we need to perform a logical AND on the w and
                # ym variables. This is done by introducing another helper
                # variable, yn.
                prob.define(('yn', reaction_id, compound),
                            types=lp.VariableType.Binary)
                yn = prob.var(('yn', reaction_id, compound))
                prob.add_linear_constraints(
                    2 * yn <= w + prob.var(('ym', reaction_id)))
                binary_cons_lhs[compound] += yn

    for compound, lhs in iteritems(binary_cons_lhs):
        if compound in blocked:
            prob.add_linear_constraints(lhs >= 1)

    # Define mass balance constraints
    massbalance_lhs = {compound: 0 for compound in model.compounds}
    for spec, value in iteritems(model.matrix):
        compound, reaction_id = spec
        massbalance_lhs[compound] += prob.var(('v', reaction_id)) * value
    for compound, lhs in iteritems(massbalance_lhs):
        # The constraint is merely >0 meaning that we have implicit sinks
        # for all compounds.
        prob.add_linear_constraints(lhs >= 0)

    # Solve
    result = prob.solve(lp.ObjectiveSense.Minimize)
    if not result:
        raise GapFillError('Non-optimal solution: {}'.format(result.status))

    def added_iter():
        for reaction_id in database_reactions:
            if result.get_value(('yd', reaction_id)) > 0.5:
                yield reaction_id

    def reversed_iter():
        for reaction_id in core:
            if result.get_value(('ym', reaction_id)) > 0.5:
                yield reaction_id

    return added_iter(), reversed_iter()



#!/usr/bin/env python
# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2014-2015  Jon Lund Steffensen <jon_steffensen@uri.edu>

import unittest

from psamm.expression.affine import Expression, Variable


class TestVariable(unittest.TestCase):
    def test_variable_init(self):
        v = Variable('x')

    def test_variable_init_long_symbol(self):
        v = Variable('xyz')

    def test_variable_init_number_symbol(self):
        v = Variable('x2')

    def test_variable_init_underscore_symbol(self):
        v = Variable('x_y')

    def test_variable_init_invalid_number(self):
        with self.assertRaises(ValueError):
            v = Variable('45x')

    def test_variable_symbol(self):
        self.assertEqual(Variable('x').symbol, 'x')

    def test_variable_simplify_returns_self(self):
        self.assertEqual(Variable('y').simplify(), Variable('y'))

    def test_variable_substitute(self):
        self.assertEqual(Variable('x').substitute(lambda v: {'x': 567}.get(v.symbol, v)), 567)

    def test_variable_substitute_unknown(self):
        self.assertEqual(Variable('x').substitute(lambda v: v), Variable('x'))

    def test_variable_substitute_multiple(self):
        self.assertEqual(Variable('x').substitute(lambda v: {'y': 123, 'x': 56}.get(v.symbol, v)), 56)

    def test_variable_add_number(self):
        self.assertEqual(Variable('x') + 1, Expression('x + 1'))

    def test_variable_radd_number(self):
        self.assertEqual(1 + Variable('x'), Expression('x + 1'))

    def test_variable_sub_number(self):
        self.assertEqual(Variable('x') - 4, Expression('x - 4'))

    def test_variable_rsub_number(self):
        self.assertEqual(4 - Variable('x'), Expression('4 - x'))

    def test_variable_mul_zero(self):
        self.assertEqual(Variable('x') * 0, 0)

    def test_variable_mul_one(self):
        self.assertEqual(Variable('x') * 1, Variable('x'))

    def test_variable_mul_number(self):
        self.assertEqual(Variable('x') * 2, Expression('2x'))

    def test_variable_rmul_number(self):
        self.assertEqual(3 * Variable('x'), Expression('3x'))

    def test_variable_div_number(self):
        self.assertEqual(Variable('x')/1, Variable('x'))

    def test_variable_neg(self):
        self.assertEqual(-Variable('x'), Expression('-x'))

    def test_variables_equals_true(self):
        self.assertEqual(Variable('x'), Variable('x'))

    def test_variables_not_equals(self):
        self.assertNotEqual(Variable('x'), Variable('y'))

    def test_variable_not_equals_number(self):
        self.assertNotEqual(Variable('x'), 5)

    def test_variables_ordered_by_symbol(self):
        var_list = sorted([Variable('b'), Variable('cd'), Variable('a'), Variable('cc')])
        self.assertEqual(var_list, [Variable('a'), Variable('b'), Variable('cc'), Variable('cd')])

    def test_variable_hash(self):
        self.assertEqual(hash(Variable('xyz')), hash(Variable('xyz')))

class TestExpression(unittest.TestCase):
    def test_expression_init_empty(self):
        e = Expression()

    def test_expression_init_with_variables(self):
        e = Expression({ Variable('a'): 1, Variable('b'): 2 })

    def test_expression_init_with_offset(self):
        e = Expression({}, 5)

    def test_expression_init_with_variables_and_offset(self):
        e = Expression({ Variable('a'): 4}, -5)

    def test_expression_init_with_zero_variables(self):
        e = Expression({ Variable('a'): 3, Variable('b'): 0})
        self.assertEqual(e, Expression({ Variable('a'): 3 }))

    def test_expression_init_with_non_variables(self):
        with self.assertRaises(ValueError):
            e = Expression({ 'a': 3 })

    def test_expression_simplify_to_number(self):
        result = Expression({}, 5).simplify()
        self.assertEqual(result, 5)
        self.assertIsInstance(result, int)

    def test_expression_simplify_to_variable(self):
        result = Expression({ Variable('x'): 1 }).simplify()
        self.assertEqual(result, Variable('x'))
        self.assertIsInstance(result, Variable)

    def test_expression_simplify_to_expression(self):
        result = Expression({ Variable('x'): 2 }).simplify()
        self.assertEqual(result, Expression('2x'))
        self.assertIsInstance(result, Expression)

    def test_expression_simplify_to_expression_with_offset(self):
        result = Expression({ Variable('x'): 1 }, 2).simplify()
        self.assertEqual(result, Expression('x + 2'))
        self.assertIsInstance(result, Expression)

    def test_expression_substitute_existing(self):
        e = Expression({ Variable('x'): 2 }, 1)
        self.assertEqual(e.substitute(lambda v: {'x': 2}.get(v.symbol, v)), 5)

    def test_expression_substitute_unknown_to_variable(self):
        e = Expression({ Variable('x'): 1 })
        self.assertEqual(e.substitute(lambda v: v), Variable('x'))

    def test_expression_substitute_unknown_to_expression(self):
        e = Expression({ Variable('x'): 2 }, 1)
        self.assertEqual(e.substitute(lambda v: v), Expression('2x + 1'))

    def test_expression_substitute_with_variable(self):
        e = Expression({ Variable('x'): 1, Variable('y'): 2 })
        self.assertEqual(e.substitute(lambda v: {'y': Variable('x')}.get(v.symbol, v)), Expression('3x'))

    def test_expression_substitute_with_expression(self):
        e = Expression({ Variable('x'): 3, Variable('y'): -2 })
        es = e.substitute(lambda v: {'x': Expression('y + 2z')}.get(v.symbol, v))
        self.assertEqual(es, Expression('y + 6z'))

    def test_expression_substitute_with_expression_is_atomic(self):
        e = Expression({ Variable('x'): 1, Variable('y'): 1 })
        es = e.substitute(lambda v: {'x': Expression('x + y'),
                                     'y': Expression('x + y') }.get(v.symbol, v))
        self.assertEqual(es, Expression('2x + 2y'))

    def test_expression_variables(self):
        e = Expression({ Variable('x'): 1, Variable('y'): 2 })
        self.assertEqual(sorted(e.variables()), [Variable('x'), Variable('y')])

    def test_expression_add_number(self):
        e = Expression({ Variable('x'): 1 })
        self.assertEqual(e + 1, Expression('x + 1'))

    def test_expression_add_other_variable(self):
        e = Expression({ Variable('x'): 1 })
        self.assertEqual(e + Variable('y'), Expression('x + y'))

    def test_expression_add_same_variable(self):
        e = Expression({ Variable('x'): 1 })
        self.assertEqual(e + Variable('x'), Expression('2x'))

    def test_expression_add_expression(self):
        e1 = Expression({ Variable('x'): 1 })
        e2 = Expression({ Variable('y'): 2 })
        self.assertEqual(e1 + e2, Expression('x + 2y'))

    def test_expression_add_sums_to_zero(self):
        e1 = Expression({ Variable('x'): 2, Variable('y'): 1 })
        e2 = Expression({ Variable('x'): -2, Variable('z'): 3 })
        self.assertEqual(e1 + e2, Expression('y + 3z'))

    def test_expression_sub_number(self):
        e = Expression({ Variable('x'): 1 })
        self.assertEqual(e - 1, Expression('x - 1'))

    def text_expression_rsub_number(self):
        e = Expression({ Variable('x'): 2 })
        self.assertEqual(4 - e, Expression('4 - 2x'))

    def test_expression_sub_other_variable(self):
        e = Expression({ Variable('x'): 1 })
        self.assertEqual(e - Variable('y'), Expression('x - y'))

    def test_expression_sub_same_variable(self):
        e = Expression({ Variable('x'): 1 })
        self.assertEqual(e - Variable('x'), 0)

    def test_expression_sub_expression(self):
        e1 = Expression({ Variable('x'): 1 })
        e2 = Expression({ Variable('y'): 2 })
        self.assertEqual(e1 - e2, Expression('x - 2y'))

    def test_expression_mul_zero(self):
        e = Expression({ Variable('x'): 3, Variable('y'): -2 }, 4)
        self.assertEqual(e * 0, 0)

    def test_expression_mul_one(self):
        e = Expression({ Variable('x'): 3, Variable('y'): -2 }, 4)
        self.assertEqual(e * 1, e)

    def test_expression_mul_number(self):
        e = Expression({ Variable('x'): 3, Variable('y'): -2 }, 4)
        self.assertEqual(e * 4, Expression('12x - 8y + 16'))

    def test_expression_div_number(self):
        e = Expression({ Variable('x'): 8, Variable('y'): -2}, 4)
        self.assertEqual(e/2, Expression('4x - y + 2'))

    def test_expression_neg(self):
        e = Expression({ Variable('x'): 3, Variable('y'): -2 }, 4)
        self.assertEqual(-e, Expression('-3x + 2y - 4'))

    def test_expression_equals_expression(self):
        e1 = Expression({ Variable('x'): 2, Variable('y'): 5 })
        e2 = Expression({ Variable('y'): 5, Variable('x'): 2 })
        self.assertEqual(e1, e2)

    def test_expression_not_equals_same_variables(self):
        e1 = Expression({ Variable('x'): 2, Variable('y'): 5 })
        e2 = Expression({ Variable('y'): 2, Variable('x'): -5 })
        self.assertNotEqual(e1, e2)

    def test_expression_not_equals_diffrent_variables(self):
        e1 = Expression({ Variable('x'): 2, Variable('y'): 5 })
        e2 = Expression({ Variable('y'): 2, Variable('x'): -5, Variable('z'): 1 })
        self.assertNotEqual(e1, e2)

    def test_expression_not_equals_with_offset(self):
        e1 = Expression({ Variable('x'): 2, Variable('y'): 5 }, 4)
        e2 = Expression({ Variable('y'): 5, Variable('x'): 2 })
        self.assertNotEqual(e1, e2)

    def test_expression_equals_variable(self):
        e = Expression({ Variable('x'): 1 })
        self.assertEqual(e, Variable('x'))

    def test_expression_with_coefficient_not_equals_variable(self):
        e = Expression({ Variable('x'): 2 })
        self.assertNotEqual(e, Variable('x'))

    def test_expression_with_offset_not_equals_variable(self):
        e = Expression({ Variable('x'): 1 }, 4)
        self.assertNotEqual(e, Variable('x'))

    def test_expression_with_multiple_not_equals_variable(self):
        e = Expression({ Variable('x'): 1, Variable('y'): 1 })
        self.assertNotEqual(e, Variable('x'))

    def test_expression_with_no_variables_equals_number(self):
        self.assertEqual(Expression({}, 3), 3)

    def test_expression_with_no_variables_not_equals_number(self):
        self.assertNotEqual(Expression({}, -3), 3)

    def test_expression_with_variables_not_equals_number(self):
        self.assertNotEqual(Expression({ Variable('x'): 1 }, 3), 3)

    def test_expression_parse(self):
        e = Expression('2x + 3')
        self.assertEqual(e, Expression({ Variable('x'): 2 }, 3))

    def test_expression_parse_number(self):
        e = Expression('1')
        self.assertEqual(e, Expression({}, 1))
        self.assertEqual(e, 1)

    def test_expression_parse_zero_offset(self):
        e = Expression('x + 4y + 0')
        self.assertEqual(e, Expression({ Variable('x'): 1, Variable('y'): 4 }))

    def test_expression_parse_multiple_occurences(self):
        e = Expression('x + 4y + 2x')
        self.assertEqual(e, Expression({ Variable('x'): 3, Variable('y'): 4 }))

    def test_expression_parse_negative_coefficient(self):
        e = Expression('-x + 4y - 3z')
        self.assertEqual(e, Expression({ Variable('x'): -1, Variable('y'): 4, Variable('z'): -3 }))

    def test_expression_parse_zero_coefficient(self):
        e = Expression('2x + 0y + 4')
        self.assertEqual(e, Expression({ Variable('x'): 2 }, 4))

    def test_expression_parse_long_variable_symbols(self):
        e = Expression('-2x1 + 5pi - 3x2')
        self.assertEqual(e, Expression({ Variable('x1'): -2, Variable('pi'): 5, Variable('x2'): -3 }))

    def test_expression_parse_no_space(self):
        e = Expression('x+3f+6h+10')
        self.assertEqual(e, Expression({ Variable('x'): 1, Variable('f'): 3, Variable('h'): 6 }, 10))

    def test_expression_parse_large_coefficient(self):
        e = Expression('x + 4200y')
        self.assertEqual(e, Expression({ Variable('x'): 1, Variable('y'): 4200 }))

    def test_expression_to_str(self):
        e = Expression({
            Variable('x1'): -2,
            Variable('pi'): 5,
            Variable('x2'): -3
        })
        self.assertEqual(str(e), '5pi - 2x1 - 3x2')


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2014-2015  Jon Lund Steffensen <jon_steffensen@uri.edu>

import unittest

from psamm.metabolicmodel import MetabolicModel
from psamm.database import DictDatabase
from psamm import fluxanalysis
from psamm.datasource.modelseed import parse_reaction
from psamm.lpsolver import generic

from six import itervalues


class TestFluxBalance(unittest.TestCase):
    def setUp(self):
        self.database = DictDatabase()
        self.database.set_reaction('rxn_1', parse_reaction('=> (2) |A|'))
        self.database.set_reaction('rxn_2', parse_reaction('|A| <=> |B|'))
        self.database.set_reaction('rxn_3', parse_reaction('|A| => |D|'))
        self.database.set_reaction('rxn_4', parse_reaction('|A| => |C|'))
        self.database.set_reaction('rxn_5', parse_reaction('|C| => |D|'))
        self.database.set_reaction('rxn_6', parse_reaction('|D| =>'))
        self.model = MetabolicModel.load_model(
            self.database, self.database.reactions)

        try:
            self.solver = generic.Solver()
        except generic.RequirementsError:
            self.skipTest('Unable to find an LP solver for tests')

    def test_flux_balance_rxn_1(self):
        fluxes = dict(fluxanalysis.flux_balance(
            self.model, 'rxn_1', tfba=False, solver=self.solver))
        self.assertAlmostEqual(fluxes['rxn_1'], 500)
        self.assertAlmostEqual(fluxes['rxn_2'], 0)
        self.assertAlmostEqual(fluxes['rxn_6'], 1000)

    def test_flux_balance_rxn_2(self):
        fluxes = dict(fluxanalysis.flux_balance(
            self.model, 'rxn_2', tfba=False, solver=self.solver))
        self.assertAlmostEqual(fluxes['rxn_2'], 0)

    def test_flux_balance_rxn_3(self):
        fluxes = dict(fluxanalysis.flux_balance(
            self.model, 'rxn_3', tfba=False, solver=self.solver))
        self.assertAlmostEqual(fluxes['rxn_1'], 500)
        self.assertAlmostEqual(fluxes['rxn_2'], 0)
        self.assertAlmostEqual(fluxes['rxn_3'], 1000)
        self.assertAlmostEqual(fluxes['rxn_6'], 1000)

    def test_flux_balance_rxn_6(self):
        fluxes = dict(fluxanalysis.flux_balance(
            self.model, 'rxn_6', tfba=False, solver=self.solver))
        self.assertAlmostEqual(fluxes['rxn_1'], 500)
        self.assertAlmostEqual(fluxes['rxn_2'], 0)
        self.assertAlmostEqual(fluxes['rxn_6'], 1000)

    def test_flux_balance_object_maximize(self):
        p = fluxanalysis.FluxBalanceProblem(self.model, self.solver)
        p.maximize('rxn_6')
        self.assertAlmostEqual(p.get_flux('rxn_1'), 500)
        self.assertAlmostEqual(p.get_flux('rxn_2'), 0)
        self.assertAlmostEqual(p.get_flux('rxn_6'), 1000)

    def test_flux_balance_object_minimize_l1(self):
        p = fluxanalysis.FluxBalanceProblem(self.model, self.solver)
        p.prob.add_linear_constraints(p.get_flux_var('rxn_6') == 1000)
        p.minimize_l1()
        self.assertAlmostEqual(p.get_flux('rxn_1'), 500)
        self.assertAlmostEqual(p.get_flux('rxn_2'), 0)
        self.assertAlmostEqual(p.get_flux('rxn_3'), 1000)
        self.assertAlmostEqual(p.get_flux('rxn_4'), 0)
        self.assertAlmostEqual(p.get_flux('rxn_5'), 0)
        self.assertAlmostEqual(p.get_flux('rxn_6'), 1000)

    def test_flux_balance_object_minimize_l1_function(self):
        fluxes = dict(fluxanalysis.flux_minimization(
            self.model, {'rxn_6': 1000}, self.solver))
        self.assertAlmostEqual(fluxes['rxn_1'], 500)
        self.assertAlmostEqual(fluxes['rxn_2'], 0)
        self.assertAlmostEqual(fluxes['rxn_3'], 1000)
        self.assertAlmostEqual(fluxes['rxn_4'], 0)
        self.assertAlmostEqual(fluxes['rxn_5'], 0)
        self.assertAlmostEqual(fluxes['rxn_6'], 1000)

    def test_flux_balance_object_max_min_l1(self):
        p = fluxanalysis.FluxBalanceProblem(self.model, self.solver)
        p.max_min_l1('rxn_6')
        self.assertAlmostEqual(p.get_flux('rxn_1'), 500)
        self.assertAlmostEqual(p.get_flux('rxn_2'), 0)
        self.assertAlmostEqual(p.get_flux('rxn_3'), 1000)
        self.assertAlmostEqual(p.get_flux('rxn_4'), 0)
        self.assertAlmostEqual(p.get_flux('rxn_5'), 0)
        self.assertAlmostEqual(p.get_flux('rxn_6'), 1000)

        # The temporary constraint on the reaction rxn_6 should go away. If
        # not, the next maximize will raise a FluxBalanceError.
        p.prob.add_linear_constraints(p.get_flux_var('rxn_1') == 10)
        p.maximize('rxn_6')
        self.assertAlmostEqual(p.get_flux('rxn_6'), 20)

    def test_flux_balance_object_max_min_l1_multiple(self):
        p = fluxanalysis.FluxBalanceProblem(self.model, self.solver)
        p.max_min_l1({'rxn_3': 1, 'rxn_4': 1, 'rxn_5': 1})
        self.assertAlmostEqual(p.get_flux('rxn_1'), 500)
        self.assertAlmostEqual(p.get_flux('rxn_2'), 0)
        self.assertAlmostEqual(p.get_flux('rxn_3'), 0)
        self.assertAlmostEqual(p.get_flux('rxn_4'), 1000)
        self.assertAlmostEqual(p.get_flux('rxn_5'), 1000)
        self.assertAlmostEqual(p.get_flux('rxn_6'), 1000)

    def test_flux_balance_infeasible(self):
        self.model.limits['rxn_6'].lower = 500
        self.model.limits['rxn_1'].upper = 10
        p = fluxanalysis.FluxBalanceProblem(self.model, self.solver)

        try:
            p.maximize('rxn_6')
        except fluxanalysis.FluxBalanceError as e:
            self.assertFalse(e.result.unbounded)
        else:
            self.fail('FluxBalanceError was not raised!')

    def test_flux_balance_unbounded(self):
        unbounded = float('-inf'), float('inf')
        self.model.limits['rxn_1'].bounds = unbounded
        self.model.limits['rxn_3'].bounds = unbounded
        self.model.limits['rxn_6'].bounds = unbounded
        p = fluxanalysis.FluxBalanceProblem(self.model, self.solver)

        if self.solver.properties['name'] == 'qsoptex':
            # QSopt_ex returns status code 100 for this example. It seems that
            # it is unable to determine whether the problem is unbounded.
            self.skipTest('Skipping because of known issue with QSopt_ex')

        try:
            p.maximize('rxn_6')
        except fluxanalysis.FluxBalanceError as e:
            self.assertTrue(e.result.unbounded)
        else:
            self.fail('FluxBalanceError was not raised!')


class TestFluxBalanceThermodynamic(unittest.TestCase):
    def setUp(self):
        self.database = DictDatabase()
        self.database.set_reaction('ex_A', parse_reaction('|A| <=>'))
        self.database.set_reaction('ex_D', parse_reaction('|D| <=>'))
        self.database.set_reaction('rxn_1', parse_reaction('|A| => |B|'))
        self.database.set_reaction('rxn_2', parse_reaction('|B| <=> |C|'))
        self.database.set_reaction('rxn_3', parse_reaction('|C| <=> |D|'))
        self.database.set_reaction('rxn_4', parse_reaction('|D| <=> |E|'))
        self.database.set_reaction('rxn_5', parse_reaction('|E| => |B|'))

        self.model = MetabolicModel.load_model(
            self.database, self.database.reactions)
        self.model.limits['ex_A'].lower = -10 # Low uptake
        self.model.limits['ex_D'].lower = 0 # No uptake

        try:
            self.solver = generic.Solver(integer=True)
        except generic.RequirementsError:
            self.skipTest('Unable to find an MILP solver for tests')

    def test_flux_balance_tfba_exchange_d(self):
        fluxes = dict(fluxanalysis.flux_balance(
            self.model, 'ex_D', tfba=True, solver=self.solver))
        self.assertAlmostEqual(fluxes['ex_A'], -10)
        self.assertAlmostEqual(fluxes['ex_D'], 10)
        self.assertAlmostEqual(fluxes['rxn_2'], 10)
        self.assertAlmostEqual(fluxes['rxn_4'], 0)
        self.assertAlmostEqual(fluxes['rxn_5'], 0)


class TestFluxVariability(unittest.TestCase):
    def setUp(self):
        self.database = DictDatabase()
        self.database.set_reaction('rxn_1', parse_reaction('=> (2) |A|'))
        self.database.set_reaction('rxn_2', parse_reaction('|A| <=> |B|'))
        self.database.set_reaction('rxn_3', parse_reaction('|A| => |D|'))
        self.database.set_reaction('rxn_4', parse_reaction('|A| => |C|'))
        self.database.set_reaction('rxn_5', parse_reaction('|C| => |D|'))
        self.database.set_reaction('rxn_6', parse_reaction('|D| =>'))
        self.database.set_reaction('rxn_7', parse_reaction('|E| => |F|'))
        self.database.set_reaction('rxn_8', parse_reaction('|F| => |E|'))

        self.model = MetabolicModel.load_model(
            self.database, self.database.reactions)
        self.model.limits['rxn_5'].upper = 100

        try:
            self.solver = generic.Solver()
        except generic.RequirementsError:
            self.skipTest('Unable to find an LP solver for tests')

    def test_flux_variability(self):
        fluxes = dict(fluxanalysis.flux_variability(
            self.model, self.model.reactions, {'rxn_6': 200},
            tfba=False, solver=self.solver))

        for bounds in itervalues(fluxes):
            self.assertEqual(len(bounds), 2)

        self.assertAlmostEqual(fluxes['rxn_1'][0], 100)

        self.assertAlmostEqual(fluxes['rxn_2'][0], 0)
        self.assertAlmostEqual(fluxes['rxn_2'][1], 0)

        self.assertAlmostEqual(fluxes['rxn_5'][0], 0)
        self.assertAlmostEqual(fluxes['rxn_5'][1], 100)

        self.assertAlmostEqual(fluxes['rxn_6'][0], 200)

        self.assertGreater(fluxes['rxn_7'][1], 0)
        self.assertGreater(fluxes['rxn_8'][1], 0)

    def test_flux_variability_unbounded(self):
        self.model.limits['rxn_7'].upper = float('inf')
        self.model.limits['rxn_8'].upper = float('inf')

        if self.solver.properties['name'] == 'qsoptex':
            # QSopt_ex returns status code 100 for this example. It seems that
            # it is unable to determine whether the problem is unbounded.
            self.skipTest('Skipping because of known issue with QSopt_ex')

        fluxes = dict(fluxanalysis.flux_variability(
            self.model, self.model.reactions, {'rxn_6': 200},
            tfba=False, solver=self.solver))

        for bounds in itervalues(fluxes):
            self.assertEqual(len(bounds), 2)

        self.assertAlmostEqual(fluxes['rxn_6'][0], 200)

        self.assertAlmostEqual(fluxes['rxn_7'][0], 0)
        self.assertEqual(fluxes['rxn_7'][1], float('inf'))

        self.assertAlmostEqual(fluxes['rxn_8'][0], 0)
        self.assertEqual(fluxes['rxn_8'][1], float('inf'))


class TestFluxVariabilityThermodynamic(unittest.TestCase):
    def setUp(self):
        self.database = DictDatabase()
        self.database.set_reaction('rxn_1', parse_reaction('=> (2) |A|'))
        self.database.set_reaction('rxn_2', parse_reaction('|A| <=> |B|'))
        self.database.set_reaction('rxn_3', parse_reaction('|A| => |D|'))
        self.database.set_reaction('rxn_4', parse_reaction('|A| => |C|'))
        self.database.set_reaction('rxn_5', parse_reaction('|C| => |D|'))
        self.database.set_reaction('rxn_6', parse_reaction('|D| =>'))
        self.database.set_reaction('rxn_7', parse_reaction('|E| => |F|'))
        self.database.set_reaction('rxn_8', parse_reaction('|F| => |E|'))

        self.model = MetabolicModel.load_model(
            self.database, self.database.reactions)
        self.model.limits['rxn_5'].upper = 100

        try:
            self.solver = generic.Solver(integer=True)
        except generic.RequirementsError:
            self.skipTest('Unable to find an MILP solver for tests')

    def test_flux_variability_thermodynamic(self):
        fluxes = dict(fluxanalysis.flux_variability(
            self.model, self.model.reactions, {'rxn_6': 200},
            tfba=True, solver=self.solver))

        self.assertAlmostEqual(fluxes['rxn_1'][0], 100)

        self.assertAlmostEqual(fluxes['rxn_2'][0], 0)
        self.assertAlmostEqual(fluxes['rxn_2'][1], 0)

        self.assertAlmostEqual(fluxes['rxn_5'][0], 0)
        self.assertAlmostEqual(fluxes['rxn_5'][1], 100)

        self.assertAlmostEqual(fluxes['rxn_6'][0], 200)

        self.assertAlmostEqual(fluxes['rxn_7'][1], 0)
        self.assertAlmostEqual(fluxes['rxn_8'][1], 0)


class TestFluxConsistency(unittest.TestCase):
    def setUp(self):
        self.database = DictDatabase()
        self.database.set_reaction('rxn_1', parse_reaction('=> (2) |A|'))
        self.database.set_reaction('rxn_2', parse_reaction('|A| <=> |B|'))
        self.database.set_reaction('rxn_3', parse_reaction('|A| => |D|'))
        self.database.set_reaction('rxn_4', parse_reaction('|A| => |C|'))
        self.database.set_reaction('rxn_5', parse_reaction('|C| => |D|'))
        self.database.set_reaction('rxn_6', parse_reaction('|D| =>'))
        self.database.set_reaction('rxn_7', parse_reaction('|E| => |F|'))
        self.database.set_reaction('rxn_8', parse_reaction('|F| => |E|'))
        self.model = MetabolicModel.load_model(
            self.database, self.database.reactions)

        try:
            self.solver = generic.Solver()
        except generic.RequirementsError:
            self.skipTest('Unable to find an LP solver for tests')

    def test_check_on_consistent(self):
        self.model.remove_reaction('rxn_2')
        core = self.model.reactions
        inconsistent = set(fluxanalysis.consistency_check(
            self.model, core, epsilon=0.001, tfba=False, solver=self.solver))
        self.assertEqual(inconsistent, set())

    def test_check_on_inconsistent(self):
        core = set(self.model.reactions)
        inconsistent = set(fluxanalysis.consistency_check(
            self.model, core, epsilon=0.001, tfba=False, solver=self.solver))
        self.assertEqual(inconsistent, {'rxn_2'})


class TestFluxConsistencyThermodynamic(unittest.TestCase):
    def setUp(self):
        self.database = DictDatabase()
        self.database.set_reaction('rxn_1', parse_reaction('=> (2) |A|'))
        self.database.set_reaction('rxn_2', parse_reaction('|A| <=> |B|'))
        self.database.set_reaction('rxn_3', parse_reaction('|A| => |D|'))
        self.database.set_reaction('rxn_4', parse_reaction('|A| => |C|'))
        self.database.set_reaction('rxn_5', parse_reaction('|C| => |D|'))
        self.database.set_reaction('rxn_6', parse_reaction('|D| =>'))
        self.database.set_reaction('rxn_7', parse_reaction('|E| => |F|'))
        self.database.set_reaction('rxn_8', parse_reaction('|F| => |E|'))
        self.model = MetabolicModel.load_model(
            self.database, self.database.reactions)

        try:
            self.solver = generic.Solver(integer=True)
        except generic.RequirementsError:
            self.skipTest('Unable to find an MILP solver for tests')

    def test_check_on_inconsistent_with_thermodynamic_constraints(self):
        self.model.remove_reaction('rxn_2')
        core = self.model.reactions
        inconsistent = set(fluxanalysis.consistency_check(
            self.model, core, epsilon=0.001, tfba=True, solver=self.solver))
        self.assertEqual(inconsistent, {'rxn_7', 'rxn_8'})


class TestFluxRandomization(unittest.TestCase):
    def setUp(self):
        self.database = DictDatabase()
        self.database.set_reaction('rxn_1', parse_reaction('=> (2) |A|'))
        self.database.set_reaction('rxn_2', parse_reaction('|A| <=> |B|'))
        self.database.set_reaction('rxn_3', parse_reaction('|A| => |D|'))
        self.database.set_reaction('rxn_4', parse_reaction('|A| => |C|'))
        self.database.set_reaction('rxn_5', parse_reaction('|C| => |D|'))
        self.database.set_reaction('rxn_6', parse_reaction('|D| =>'))
        self.database.set_reaction('rxn_7', parse_reaction('|E| => |F|'))
        self.database.set_reaction('rxn_8', parse_reaction('|F| => |E|'))
        self.model = MetabolicModel.load_model(
            self.database, self.database.reactions)

        try:
            self.solver = generic.Solver()
        except generic.RequirementsError:
            self.skipTest('Unable to find an LP solver for tests')

    def test_flux_randomization(self):
        fluxes = dict(fluxanalysis.flux_randomization(
            self.model, {'rxn_6': 1000}, False, self.solver))
        self.assertAlmostEqual(fluxes['rxn_1'], 500)
        self.assertAlmostEqual(fluxes['rxn_2'], 0)
        self.assertAlmostEqual(fluxes['rxn_3'] + fluxes['rxn_4'], 1000)
        self.assertAlmostEqual(fluxes['rxn_6'], 1000)

        # Cycle
        self.assertGreaterEqual(fluxes['rxn_7'], 0)
        self.assertGreaterEqual(fluxes['rxn_8'], 0)
        self.assertLessEqual(fluxes['rxn_7'], 1000)
        self.assertLessEqual(fluxes['rxn_8'], 1000)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2014-2015  Jon Lund Steffensen <jon_steffensen@uri.edu>

import unittest

from psamm.metabolicmodel import MetabolicModel
from psamm.database import DictDatabase
from psamm import fastcore
from psamm.datasource.modelseed import parse_reaction
from psamm.lpsolver import generic


class TestFastcoreSimpleVlassisModel(unittest.TestCase):
    """Test fastcore using the simple model in [Vlassis14]_."""

    def setUp(self):
        # TODO use mock model instead of actual model
        self.database = DictDatabase()
        self.database.set_reaction('rxn_1', parse_reaction('=> (2) |A|'))
        self.database.set_reaction('rxn_2', parse_reaction('|A| <=> |B|'))
        self.database.set_reaction('rxn_3', parse_reaction('|A| => |D|'))
        self.database.set_reaction('rxn_4', parse_reaction('|A| => |C|'))
        self.database.set_reaction('rxn_5', parse_reaction('|C| => |D|'))
        self.database.set_reaction('rxn_6', parse_reaction('|D| =>'))
        self.model = MetabolicModel.load_model(
            self.database, self.database.reactions)

        try:
            self.solver = generic.Solver()
        except generic.RequirementsError:
            self.skipTest('Unable to find an LP solver for tests')

    def test_lp10(self):
        p = fastcore.FastcoreProblem(self.model, self.solver, epsilon=0.001)
        scaling = 1000
        p.lp10({'rxn_6'}, {'rxn_1', 'rxn_3', 'rxn_4', 'rxn_5'})
        supp = set(reaction_id for reaction_id in self.model.reactions
                   if abs(p.get_flux(reaction_id)) >= 0.999 * 0.001 / scaling)
        self.assertEqual(supp, { 'rxn_1', 'rxn_3', 'rxn_6' })

    def test_lp10_weighted(self):
        p = fastcore.FastcoreProblem(self.model, self.solver, epsilon=0.001)
        scaling = 1000
        weights = {'rxn_3': 1}
        p.lp10({'rxn_6'}, {'rxn_1', 'rxn_3', 'rxn_4', 'rxn_5'},
               weights=weights)
        supp = set(reaction_id for reaction_id in self.model.reactions
                   if abs(p.get_flux(reaction_id)) >= 0.999 * 0.001 / scaling)
        self.assertEqual(supp, {'rxn_1', 'rxn_3', 'rxn_6'})

        weights = {'rxn_3': 3}
        p.lp10({'rxn_6'}, {'rxn_1', 'rxn_3', 'rxn_4', 'rxn_5'},
               weights=weights)
        supp = set(reaction_id for reaction_id in self.model.reactions
                   if abs(p.get_flux(reaction_id)) >= 0.999 * 0.001 / scaling)
        self.assertEqual(supp, {'rxn_1', 'rxn_4', 'rxn_5', 'rxn_6'})

    def test_lp7(self):
        p = fastcore.FastcoreProblem(self.model, self.solver, epsilon=0.001)
        p.lp7(set(self.model.reactions))
        supp = set(reaction_id for reaction_id in self.model.reactions
                   if p.get_flux(reaction_id) >= 0.001*0.999)
        self.assertEqual(supp, {'rxn_1', 'rxn_3', 'rxn_4', 'rxn_5', 'rxn_6'})

        p.lp7({'rxn_5'})
        supp = set(reaction_id for reaction_id in self.model.reactions
                   if p.get_flux(reaction_id) >= 0.001*0.999)
        # Test that the support contains at least the given reactions
        self.assertLessEqual({'rxn_4', 'rxn_5', 'rxn_6'}, supp)

    def test_find_sparse_mode_singleton(self):
        p = fastcore.FastcoreProblem(self.model, self.solver, epsilon=0.001)
        core = {'rxn_1'}
        mode = set(p.find_sparse_mode(
            core, set(self.model.reactions) - core, scaling=1e3))
        self.assertEqual(mode, {'rxn_1', 'rxn_3', 'rxn_6'})

        core = {'rxn_2'}
        mode = set(p.find_sparse_mode(
            core, set(self.model.reactions) - core, scaling=1e3))
        self.assertEqual(mode, set())

        core = {'rxn_3'}
        mode = set(p.find_sparse_mode(
            core, set(self.model.reactions) - core, scaling=1e3))
        self.assertEqual(mode, {'rxn_1', 'rxn_3', 'rxn_6'})

        core = {'rxn_4'}
        mode = set(p.find_sparse_mode(
            core, set(self.model.reactions) - core, scaling=1e3))
        self.assertEqual(mode, {'rxn_1', 'rxn_4', 'rxn_5', 'rxn_6'})

        core = {'rxn_5'}
        mode = set(p.find_sparse_mode(
            core, set(self.model.reactions) - core, scaling=1e3))
        self.assertEqual(mode, {'rxn_1', 'rxn_4', 'rxn_5', 'rxn_6'})

        core = {'rxn_6'}
        mode = set(p.find_sparse_mode(
            core, set(self.model.reactions) - core, scaling=1e3))
        self.assertEqual(mode, {'rxn_1', 'rxn_3', 'rxn_6'})

    def test_find_sparse_mode_weighted(self):
        p = fastcore.FastcoreProblem(self.model, self.solver, epsilon=0.001)
        core = {'rxn_1'}
        weights = {'rxn_3': 1}
        mode = set(p.find_sparse_mode(
            core, set(self.model.reactions) - core,
            scaling=1e3, weights=weights))
        self.assertEqual(mode, {'rxn_1', 'rxn_3', 'rxn_6'})

        weights = {'rxn_3': 3}
        mode = set(p.find_sparse_mode(
            core, set(self.model.reactions) - core,
            scaling=1e3, weights=weights))
        self.assertEqual(mode, {'rxn_1', 'rxn_4', 'rxn_5', 'rxn_6'})

    def test_fastcc_inconsistent(self):
        self.assertEqual(
            set(fastcore.fastcc(self.model, 0.001, solver=self.solver)),
            {'rxn_2'})

    def test_fastcc_is_consistent_on_inconsistent(self):
        self.assertFalse(fastcore.fastcc_is_consistent(
            self.model, 0.001, solver=self.solver))

    def test_fastcc_is_consistent_on_consistent(self):
        self.model.remove_reaction('rxn_2')
        self.assertTrue(fastcore.fastcc_is_consistent(
            self.model, 0.001, solver=self.solver))

    def test_fastcc_consistent_subset(self):
        self.assertEqual(fastcore.fastcc_consistent_subset(
            self.model, 0.001, solver=self.solver),
            set(['rxn_1', 'rxn_3', 'rxn_4', 'rxn_5', 'rxn_6']))

    def test_fastcore_global_inconsistent(self):
        self.database.set_reaction('rxn_7', parse_reaction('|E| <=>'))
        self.model.add_reaction('rxn_7')
        with self.assertRaises(fastcore.FastcoreError):
            fastcore.fastcore(self.model, {'rxn_7'}, 0.001,
                              solver=self.solver)


class TestFastcoreTinyBiomassModel(unittest.TestCase):
    """Test fastcore using a model with tiny values in biomass reaction

    This model is consistent mathematically since there is a flux solution
    within the flux bounds. However, the numerical nature of the fastcore
    algorithm requires an epsilon-parameter indicating the minimum flux that
    is considered non-zero. For this reason, some models with reactions where
    tiny stoichiometric values appear can be seen as inconsistent by
    fastcore.

    In this particular model, rxn_2 can take a maximum flux of 1000. At the
    same time rxn_1 will have to take a flux of 1e-3. This is the maximum
    possible flux for rxn_1 so running fastcore with an epsilon larger than
    1e-3 will indicate that the model is not consistent.
    """

    def setUp(self):
        # TODO use mock model instead of actual model
        self.database = DictDatabase()
        self.database.set_reaction('rxn_1', parse_reaction('=> |A|'))
        self.database.set_reaction('rxn_2',
            parse_reaction('(0.000001) |A| =>'))
        self.model = MetabolicModel.load_model(
            self.database, self.database.reactions)

        try:
            self.solver = generic.Solver()
        except generic.RequirementsError:
            self.skipTest('Unable to find an LP solver for tests')

    def test_fastcc_is_consistent(self):
        self.assertTrue(fastcore.fastcc_is_consistent(
            self.model, 0.0001, solver=self.solver))

    def test_fastcc_is_consistent_high_epsilon(self):
        self.assertFalse(fastcore.fastcc_is_consistent(
            self.model, 0.1, solver=self.solver))

    def test_fastcore_induced_model(self):
        core = {'rxn_2'}
        self.assertEqual(
            set(fastcore.fastcore(
                self.model, core, 0.0001, scaling=1e7, solver=self.solver)),
            {'rxn_1', 'rxn_2'})

    def test_fastcore_induced_model_high_epsilon(self):
        core = {'rxn_2'}
        self.assertEqual(
            set(fastcore.fastcore(
                self.model, core, 0.1, scaling=1e7, solver=self.solver)),
            {'rxn_1', 'rxn_2'})


class TestFlippingModel(unittest.TestCase):
    """Test fastcore on a model that has to flip"""

    def setUp(self):
        # TODO use mock model instead of actual model
        self.database = DictDatabase()
        self.database.set_reaction('rxn_1', parse_reaction('|A| <=>'))
        self.database.set_reaction('rxn_2', parse_reaction('|A| <=> |B|'))
        self.database.set_reaction('rxn_3', parse_reaction('|C| <=> |B|'))
        self.database.set_reaction('rxn_4', parse_reaction('|C| <=>'))
        self.model = MetabolicModel.load_model(
            self.database, self.database.reactions)

        try:
            self.solver = generic.Solver()
        except generic.RequirementsError:
            self.skipTest('Unable to find an LP solver for tests')

    def test_fastcore_induced_model(self):
        core = {'rxn_2', 'rxn_3'}
        self.assertEqual(set(
            fastcore.fastcore(self.model, core, 0.001, solver=self.solver)),
            {'rxn_1', 'rxn_2', 'rxn_3', 'rxn_4'})


if __name__ == '__main__':
    unittest.main()



# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2015  Keith Dufault-Thompson <keitht547@my.uri.edu>

import logging

from ..command import Command

logger = logging.getLogger(__name__)


class ExportTableCommand(Command):
    """Export parts of the model as tab separated tables.

    This command can export various parts of the model as a TSV table.
    The output is written to standard output. The type argument is used
    to determine which part of the model to export:

    - reactions: Export reactions and reaction metadata
    - compounds: Export compounds and compound metadata
    - medium: Export the list of exchange compounds/reactions
    - limits: Export list of internal flux limits
    - metadata: Export general model metadata
    """

    @classmethod
    def init_parser(cls, parser):
        parser.add_argument(
            'export', metavar='export_type',
            choices=['reactions', 'compounds', 'medium', 'limits', 'metadata'],
            help='Type of model data to export')

    def run(self):
        export_type = self._args.export
        if export_type == 'reactions':
            self._reaction_export()
        elif export_type == 'compounds':
            self._compound_export()
        elif export_type == 'medium':
            self._media_export()
        elif export_type == 'limits':
            self._limits_export()
        elif export_type == 'metadata':
            self._metadata_export()

    def _reaction_export(self):
        property_set = set()
        for j in self._model.parse_reactions():
            property_set.update(j.properties)

        property_list_sorted = sorted(
            property_set, key=lambda x: (x != 'id', x != 'equation', x))

        print('\t'.join([str(x) for x in property_list_sorted]))
        for i in self._model.parse_reactions():
            print('\t'.join(str(i.properties.get(property))
                            for property in property_list_sorted))

    def _compound_export(self):
        compound_set = set()
        for compound in self._model.parse_compounds():
            compound_set.update(compound.properties)

        compound_list_sorted = sorted(
            compound_set, key=lambda x: (x != 'id', x != 'name', x))

        print('\t'.join([str(x) for x in compound_list_sorted]))
        for compound in self._model.parse_compounds():
            print('\t'.join(str(compound.properties.get(property))
                            for property in compound_list_sorted))

    def _media_export(self):
        print('{}\t{}\t{}\t{}'.format('Compound ID', 'Reaction ID',
                                      'Lower Limit', 'Upper Limit'))
        default_flux = self._model.get_default_flux_limit()
        if default_flux is None:
            default_flux = 1000

        for compound, reaction, lower, upper in self._model.parse_medium():
            if lower is None:
                lower = -1 * default_flux

            if upper is None:
                upper = default_flux

            print('{}\t{}\t{}\t{}'.format(compound, reaction, lower, upper))

    def _limits_export(self):
        print('{}\t{}\t{}'.format(
            'Reaction ID', 'Lower Limits', 'Upper Limits'))
        for reaction, lower, upper in self._model.parse_limits():
            print('{}\t{}\t{}'.format(reaction, lower, upper))

    def _metadata_export(self):
        print('Model Name: {}'.format(self._model.get_name()))
        print('Biomass Reaction: {}'.format(
            self._model.get_biomass_reaction()))
        print('Default Flux Limits: {}'.format(
            self._model.get_default_flux_limit()))



# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2014-2015  Jon Lund Steffensen <jon_steffensen@uri.edu>
# Copyright 2015  Keith Dufault-Thompson <keitht547@my.uri.edu>

import time
import random
import logging

from ..command import Command, SolverCommandMixin, CommandError
from .. import fluxanalysis, util
from ..expression import boolean

from six import string_types

logger = logging.getLogger(__name__)


class RandomSparseNetworkCommand(SolverCommandMixin, Command):
    """Find a random minimal network of model reactions.

    Given a reaction to optimize and a threshold, delete reactions randomly
    until the flux of the reaction to optimize falls under the threshold.
    Keep deleting reactions until no more reactions can be deleted. By default
    this uses standard FBA (not tFBA). Since the internal fluxes are irrelevant
    the FBA and tFBA are equivalent for this purpose.

    The threshold can be specified as an absolute flux (e.g. '1.23') or a
    relative flux of the full model flux (e.g. '40.5%').
    """

    @classmethod
    def init_parser(cls, parser):
        parser.add_argument(
            '--tfba', help='Enable thermodynamic constraints on FBA',
            action='store_true', default=False)
        parser.add_argument('--objective', help='Reaction flux to maximize')
        parser.add_argument(
            'threshold', help='Threshold of max reaction flux',
            type=util.MaybeRelative)
        parser.add_argument(
            '--type', help='Type of deletion to perform',
            choices=['reactions', 'exchange', 'genes'], type=str,
            required=True)
        super(RandomSparseNetworkCommand, cls).init_parser(parser)

    def run(self):
        if self._args.objective is not None:
            reaction = self._args.objective
        else:
            reaction = self._model.get_biomass_reaction()
            if reaction is None:
                raise CommandError('The biomass reaction was not specified')

        if not self._mm.has_reaction(reaction):
            raise CommandError('Specified reaction is not in model: {}'.format(
                reaction))

        if not self._args.tfba:
            solver = self._get_solver()
        else:
            solver = self._get_solver(integer=True)

        p = fluxanalysis.FluxBalanceProblem(self._mm, solver)

        if self._args.tfba:
            p.add_thermodynamic()

        threshold = self._args.threshold
        if threshold.relative:
            p.maximize(reaction)
            threshold.reference = p.get_flux(reaction)

        flux_threshold = float(threshold)

        logger.info('Flux threshold for {} is {}'.format(
            reaction, flux_threshold))

        if self._args.type in ('reactions', 'exchange'):
            self._delete_reactions(p, reaction, flux_threshold)
        elif self._args.type == 'genes':
            self._delete_genes(p, reaction, flux_threshold)

    def _delete_reactions(self, prob, obj_reaction, flux_threshold):
        essential = {obj_reaction}
        if self._args.type == 'exchange':
            reactions = set()
            for reaction_id in self._mm.reactions:
                if self._mm.is_exchange(reaction_id):
                    reactions.add(reaction_id)
        else:
            reactions = set(self._mm.reactions)
        test_set = reactions - essential
        deleted = set()

        start_time = time.time()

        while len(test_set) > 0:
            testing_reaction = random.sample(test_set, 1)[0]
            test_set.remove(testing_reaction)

            flux_var = prob.get_flux_var(testing_reaction)
            c, = prob.prob.add_linear_constraints(flux_var == 0)

            logger.info('Trying FBA without reaction {}...'.format(
                testing_reaction))

            try:
                prob.maximize(obj_reaction)
            except fluxanalysis.FluxBalanceError:
                logger.info(
                    'FBA is infeasible, marking {} as essential'.format(
                        testing_reaction))
                c.delete()
                essential.add(testing_reaction)
                continue

            logger.debug('Reaction {} has flux {}'.format(
                obj_reaction, prob.get_flux(obj_reaction)))

            if prob.get_flux(obj_reaction) < flux_threshold:
                c.delete()
                essential.add(testing_reaction)
                logger.info('Reaction {} was essential'.format(
                    testing_reaction))
            else:
                deleted.add(testing_reaction)
                logger.info('Reaction {} was deleted'.format(testing_reaction))

        logger.info('Solving took {:.2f} seconds'.format(
            time.time() - start_time))

        for reaction_id in sorted(reactions):
            value = 0 if reaction_id in deleted else 1
            print('{}\t{}'.format(reaction_id, value))

        logger.info('Essential reactions: {}/{}'.format(
            len(essential), len(reactions)))

    def _delete_genes(self, prob, obj_reaction, flux_threshold):
        genes = set()
        gene_assoc = {}
        for reaction in self._model.parse_reactions():
            assoc = None
            if reaction.genes is None:
                continue
            elif isinstance(reaction.genes, string_types):
                assoc = boolean.Expression(reaction.genes)
            else:
                variables = [boolean.Variable(g) for g in reaction.genes]
                assoc = boolean.Expression(boolean.And(*variables))
            genes.update(v.symbol for v in assoc.variables)
            gene_assoc[reaction.id] = assoc

        essential = set()
        deleted = set()
        test_set = set(genes)
        reactions = set(self._mm.reactions)

        start_time = time.time()

        while len(test_set) > 0:
            testing_gene = random.sample(test_set, 1)[0]
            test_set.remove(testing_gene)
            new_gene_assoc = {}
            deleted_reactions = set()

            logger.info('Trying model without gene {}...'.format(
                testing_gene))

            for reaction in reactions:
                if reaction not in gene_assoc:
                    continue
                assoc = gene_assoc[reaction]
                if boolean.Variable(testing_gene) in assoc.variables:
                    new_assoc = assoc.substitute(
                        lambda v: v if v.symbol != testing_gene else False)
                    if new_assoc.has_value() and not new_assoc.value:
                        deleted_reactions.add(reaction)
                    else:
                        new_gene_assoc[reaction] = new_assoc
                else:
                    new_gene_assoc[reaction] = assoc

            if obj_reaction in deleted_reactions:
                logger.info(
                    'Marking gene {} as essential because the objective'
                    ' reaction depends on this gene...'.format(testing_gene))
                essential.add(testing_gene)
                continue

            if len(deleted_reactions) == 0:
                logger.info(
                    'No reactions were removed when gene {}'
                    ' was deleted'.format(testing_gene))
                deleted.add(testing_gene)
                gene_assoc = new_gene_assoc
                continue

            logger.info(
                'Reactions removed when gene {} is deleted: {}'.format(
                    testing_gene, len(deleted_reactions)))
            logger.info('Deleted reactions: {}'.format(
                ', '.join(deleted_reactions)))

            constraints = []
            for reaction in deleted_reactions:
                flux_var = prob.get_flux_var(reaction)
                c, = prob.prob.add_linear_constraints(flux_var == 0)
                constraints.append(c)

            try:
                prob.maximize(obj_reaction)
            except fluxanalysis.FluxBalanceError:
                logger.info(
                    'FBA is infeasible, marking {} as essential'.format(
                        testing_gene))
                for c in constraints:
                    c.delete()
                essential.add(testing_gene)
                continue

            logger.debug('Reaction {} has flux {}'.format(
                obj_reaction, prob.get_flux(obj_reaction)))

            if prob.get_flux(obj_reaction) < flux_threshold:
                for c in constraints:
                    c.delete()
                essential.add(testing_gene)
                logger.info('Gene {} was essential'.format(testing_gene))
            else:
                deleted.add(testing_gene)
                reactions.difference_update(deleted_reactions)
                gene_assoc = new_gene_assoc
                logger.info('Gene {} was deleted'.format(testing_gene))

        logger.info('Solving took {:.2f} seconds'.format(
            time.time() - start_time))

        for gene_id in sorted(genes):
            value = 0 if gene_id in deleted else 1
            print('{}\t{}'.format(gene_id, value))

        logger.info('Essential genes: {}/{}'.format(
            len(essential), len(genes)))

        logger.info('Reactions in minimal network: {}'.format(
            ', '.join(sorted(reactions))))



# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2014-2015  Jon Lund Steffensen <jon_steffensen@uri.edu>

"""Module related to loading ModelSEED database files."""

import csv
import re
from decimal import Decimal

from ..reaction import Reaction, Compound
from .context import FileMark


class ParseError(Exception):
    """Exception used to signal errors while parsing"""


def decode_name(s):
    """Decode names in ModelSEED files"""
    # Some names contain XML-like entity codes
    return re.sub(r'&#(\d+);', lambda x: chr(int(x.group(1))), s)


class CompoundEntry(object):
    """Representation of entry in a ModelSEED compound table"""

    def __init__(self, id, names, formula, filemark=None):
        self._id = id
        self._properties = {
            'id': self._id,
            'names': list(names),
            'formula': formula
        }

        # Find shortest name
        # Usually the best name is the shortest one,
        # but in the case of ties, the best one is
        # usually the last one to appear in the list.
        # Since the sort is stable we can obtain this
        # by first reversing the list, then sorting
        # by length.
        names = self._properties['names']
        self._properties['name'] = None
        if len(self._properties['names']) > 0:
            name = sorted(reversed(names), key=lambda x: len(x))[0]
            self._properties['name'] = name

        self._filemark = filemark

    @property
    def id(self):
        return self._id

    @property
    def name(self):
        return self._properties.get('name')

    @property
    def names(self):
        return iter(self._properties.get('names'))

    @property
    def formula(self):
        return self._properties.get('formula')

    @property
    def properties(self):
        return dict(self._properties)

    @property
    def filemark(self):
        return self._filemark


def parse_compound_file(f, context=None):
    """Iterate over the compound entries in the given file"""

    f.readline()  # Skip header
    for lineno, row in enumerate(csv.reader(f, delimiter='\t')):
        compound_id, names, formula = row[:3]
        names = (decode_name(name) for name in names.split(',<br>'))

        # ModelSEED sometimes uses an asterisk and number at
        # the end of formulas. This seems to have a similar
        # meaning as '(...)n'.
        m = re.match(r'^(.*)\*(\d*)$', formula)
        if m is not None:
            if m.group(2) != '':
                formula = '({}){}'.format(m.group(1), m.group(2))
            else:
                formula = '({})n'.format(m.group(1))

        formula = formula.strip()
        if formula == '' or formula == 'noformula':
            formula = None

        mark = FileMark(context, lineno, 0)
        yield CompoundEntry(compound_id, names, formula, filemark=mark)


def parse_reaction(s):
    """Parse a ModelSEED reaction

    This parser is based on the grammer.::

        <reaction>     ::= <comp-list> ' ' <reaction-dir> ' ' <comp-list>
        <reaction-dir> ::= '<=' | '<=>' | '=>' | '?' | ''
        <comp-list>    ::= '' | <compound> | <compound> ' + ' <comp-list>
        <compound>     ::= <comp-count> ' ' <comp-spec> | <comp-spec>
        <comp-count>   ::= '(' <comp-number> ')' | <comp-number>
        <comp-number>  ::= <decimal>
        <comp-spec>    ::= '|' <comp-id> '|' | 'cpd' <cpd-id> | 'cdp' <cpd-id>
        <comp-id>      ::= <comp-name> '[' <comp-compart> ']' | <comp-name>
        <comp-compart> ::= <alpha>
        <comp-name>    ::= <any characters other than "|">
        <cpd-id>       ::= <five digits>
    """

    def tokenize(s):
        """Return tokens of reaction string"""
        s = s.lstrip()
        token = ''
        barquote = False
        for c in s:
            if c.isspace() and not barquote:
                yield token
                token = ''
                continue

            token += c

            if c == '|':
                barquote = not barquote

        if token != '':
            yield token

    def parse_compound_number(number):
        """Parse compound number

        Return plain int if possible, otherwise use Decimal."""

        d = Decimal(number)
        if d % 1 == 0:
            return int(d)
        return d

    def parse_compound_count(count):
        """Parse compound count"""

        m = re.match(r'^\((.+)\)|(.+)$', count)
        if not m:
            raise ParseError(
                'Unable to parse compound count: {}'.format(count))

        number = m.group(1) if m.group(1) is not None else m.group(2)
        return parse_compound_number(number)

    def parse_compound_name(name):
        """Parse compound name"""

        m = re.match(r'^\|(.+)\||(cdp\d+.*)|(cpd\d+.*)$', name)
        if not m:
            raise ParseError('Unable to parse compound name: {}'.format(name))

        # Return first matching group
        for g in m.groups():
            if g is not None:
                return g

        return None

    def parse_compound(cmpd):
        """Parse compound"""

        count = 1
        if len(cmpd) == 2:
            count = parse_compound_count(cmpd[0])
            name = parse_compound_name(cmpd[1])
        elif len(cmpd) == 1:
            name = parse_compound_name(cmpd[0])
        else:
            raise ParseError(
                'Unexpected number of tokens in compound: {}'.format(cmpd))

        compartment = None
        m = re.match(r'^(.+?)\[(.+)\]$', name)
        if m is not None:
            name = m.group(1)
            compartment = m.group(2)

        return Compound(name, compartment=compartment), count

    def parse_compound_list(cmpds):
        """Parse a list of compounds"""

        if len(cmpds) == 0:
            return

        cmpd = []
        for t in cmpds:
            if t == '+':
                yield parse_compound(cmpd)
                cmpd = []
            else:
                cmpd.append(t)

        if len(cmpd) == 0:
            raise ParseError('Expected compound in compound list')

        yield parse_compound(cmpd)

    tokens = list(tokenize(s))
    direction = None
    for i, t in enumerate(tokens):
        if t in ('<=', '<=>', '=>', '?', ''):
            direction = t
            left = tokens[:i]
            right = tokens[i+1:]

    if direction is None:
        raise ParseError('Failed to parse reaction: {}'.format(tokens))

    if direction in ('', '?'):
        direction = Reaction.Bidir

    return Reaction(direction, parse_compound_list(left),
                    parse_compound_list(right))


def format_reaction(reaction):
    """Format reaction as string

    converting parsed reactions
    back to string represetation using this module only a subset of the grammar
    is used.::

        <reaction>     ::= <comp-list> ' ' <reaction-dir> ' ' <comp-list>
        <reaction-dir> ::= '<=>' | '=>' | '?'
        <comp-list>    ::= '' | <compound> | <compound> ' + ' <comp-list>
        <compound>     ::= <comp-count> ' ' <comp-spec> | <comp-spec>
        <comp-count>   ::= '(' <decimal> ')'
        <comp-spec>    ::= '|' <comp-name> '|' | '|' <comp-name>
                           '[' <comp-compart> ']' '|'
        <comp-compart> ::= <alpha>
        <comp-name>    ::= <any characters other than "|">
    """

    # Define helper functions
    def format_compound(compound, count):
        """Format compound"""
        cpdspec = str(compound)
        if count != 1:
            return '({}) |{}|'.format(count, cpdspec)
        return '|{}|'.format(cpdspec)

    def format_compound_list(cmpds):
        """Format compound list"""
        return ' + '.join(format_compound(compound, count)
                          for compound, count in cmpds)

    return '{} {} {}'.format(
        format_compound_list(reaction.left),
        '?' if reaction.direction == '' else reaction.direction,
        format_compound_list(reaction.right))



#!/usr/bin/env python
# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2014-2015  Jon Lund Steffensen <jon_steffensen@uri.edu>

import unittest

from psamm.datasource import kegg
from psamm.reaction import Reaction, Compound
from psamm.expression.affine import Expression

from six import StringIO


class MockEntryClass(object):
    def __init__(self, props, filemark=None):
        self.props = props
        self.filemark = filemark


class TestKEGGEntryParser(unittest.TestCase):
    def setUp(self):
        self.f = StringIO('\n'.join([
            'ENTRY     ID001',
            'NAME      Test entry',
            'PROPERTY  This is a multi-line',
            '          property!',
            '///',
            'ENTRY     ID002',
            'NAME      Another entry',
            'PROPERTY  Single line property',
            'REFS      ref1: abcdef',
            '          ref2: defghi',
            '///'
        ]))

    def test_parse_entries(self):
        entries = list(kegg._parse_kegg_entries(self.f, MockEntryClass))
        self.assertEqual(len(entries), 2)

        self.assertEqual(entries[0].props, {
            'entry': ['ID001'],
            'name': ['Test entry'],
            'property': ['This is a multi-line', 'property!']
        })

        self.assertEqual(entries[1].props, {
            'entry': ['ID002'],
            'name': ['Another entry'],
            'property': ['Single line property'],
            'refs': [
                'ref1: abcdef',
                'ref2: defghi'
            ]
        })


class TestKEGGCompoundEntry(unittest.TestCase):
    def test_minimal_compound_entry(self):
        c = kegg.CompoundEntry({
            'entry': ['C00001        Compound']
        })

        self.assertEqual(c.id, 'C00001')
        self.assertIsNone(c.name)
        self.assertEqual(list(c.names), [])
        self.assertIsNone(c.formula)
        self.assertIsNone(c.exact_mass)
        self.assertIsNone(c.mol_weight)
        self.assertIsNone(c.comment)
        self.assertIsNone(c.filemark)

    def test_complete_compound_entry(self):
        c = kegg.CompoundEntry({
            'entry': ['C00001        Compound'],
            'name': ['H2O;', 'Water'],
            'reaction': ['R00001 R00002', 'R00003'],
            'enzyme': ['1.2.3.4 2.3.4.5', '7.6.50.4 2.1.-,-'],
            'formula': ['H2O'],
            'exact_mass': ['18.01'],
            'mol_weight': ['18.01'],
            'pathway': [
                'map00001  First pathway',
                'map00002  Second pathway'
            ],
            'dblinks': [
                'CAS: 12345',
                'ChEBI: B2345'
            ],
            'comment': ['This information is purely for testing!']
        })

        self.assertEqual(c.id, 'C00001')
        self.assertEqual(c.name, 'H2O')
        self.assertEqual(list(c.names), ['H2O', 'Water'])
        self.assertEqual(list(c.reactions), ['R00001', 'R00002', 'R00003'])
        self.assertEqual(list(c.enzymes), [
            '1.2.3.4', '2.3.4.5', '7.6.50.4', '2.1.-,-'])
        self.assertEqual(c.formula, 'H2O')
        self.assertAlmostEqual(c.exact_mass, 18.01)
        self.assertAlmostEqual(c.mol_weight, 18.01)
        self.assertEqual(list(c.pathways), [
            ('map00001', 'First pathway'),
            ('map00002', 'Second pathway')
        ])
        self.assertEqual(list(c.dblinks), [
            ('CAS', '12345'), ('ChEBI', 'B2345')
        ])
        self.assertEqual(c.comment, 'This information is purely for testing!')


class TestKEGGReactionParser(unittest.TestCase):
    def test_kegg_parse(self):
        r = kegg.parse_reaction('C00013 + C00001 <=> 2 C00009')
        self.assertEqual(r, Reaction(Reaction.Bidir, [(Compound('C00013'), 1), (Compound('C00001'), 1)],
                                      [(Compound('C00009'), 2)]))

    def test_kegg_parse_with_count_expression(self):
        r = kegg.parse_reaction('2n C00404 + n C00001 <=> (n+1) C02174')
        self.assertEqual(r, Reaction(Reaction.Bidir, [(Compound('C00404'), Expression('2n')),
                                                       (Compound('C00001'), Expression('n'))],
                                      [(Compound('C02174'), Expression('n+1'))]))

    def test_kegg_parse_with_compound_argument(self):
        r = kegg.parse_reaction('C00039(n) <=> C00013 + C00039(n+1)')
        self.assertEqual(r, Reaction(Reaction.Bidir, [(Compound('C00039', arguments=[Expression('n')]), 1)],
                                      [(Compound('C00013'), 1), (Compound('C00039', arguments=[Expression('n+1')]), 1)]))


if __name__ == '__main__':
    unittest.main()



# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2014-2015  Jon Lund Steffensen <jon_steffensen@uri.edu>

"""Linear programming solver using Cplex."""

from __future__ import absolute_import

import logging
from itertools import repeat, count
import numbers

from six.moves import zip
import cplex as cp

from .lp import Solver as BaseSolver
from .lp import Constraint as BaseConstraint
from .lp import Problem as BaseProblem
from .lp import Result as BaseResult
from .lp import (Expression, Relation, ObjectiveSense, VariableType,
                 InvalidResultError)
from ..util import LoggerFile

# Module-level logging
logger = logging.getLogger(__name__)

_INF = float('inf')


class Solver(BaseSolver):
    """Represents an LP-solver using Cplex"""

    def create_problem(self, **kwargs):
        """Create a new LP-problem using the solver"""
        return Problem(**kwargs)


class Problem(BaseProblem):
    """Represents an LP-problem of a cplex.Solver"""

    VARTYPE_MAP = {
        VariableType.Continuous: 'C',
        VariableType.Binary: 'B',
        VariableType.Integer: 'I'
    }

    CONSTR_SENSE_MAP = {
        Relation.Equals: 'E',
        Relation.Greater: 'G',
        Relation.Less: 'L'
    }

    def __init__(self, **kwargs):
        self._cp = cp.Cplex()

        # Set up output to go to logging streams
        log_stream = LoggerFile(logger, logging.DEBUG)
        warning_stream = LoggerFile(logger, logging.WARNING)
        error_stream = LoggerFile(logger, logging.ERROR)

        self._cp.set_log_stream(log_stream)
        self._cp.set_results_stream(log_stream)
        self._cp.set_warning_stream(warning_stream)
        self._cp.set_error_stream(error_stream)

        # Set feasibility tolerance. By default, we decrease it to 1e-9.
        feasibility_tolerance = kwargs.get('feasibility_tolerance', 1e-9)
        logger.info('Setting feasibility tolerance to {!r}'.format(
            feasibility_tolerance))
        self._cp.parameters.simplex.tolerances.feasibility.set(
            feasibility_tolerance)

        # Set number of threads
        if 'threads' in kwargs:
            logger.info('Setting threads to {!r}'.format(kwargs['threads']))
            self._cp.parameters.threads.set(kwargs['threads'])

        self._cp.parameters.emphasis.numerical.set(True)

        self._variables = {}
        self._var_names = (i for i in count(0))
        self._constr_names = ('c'+str(i) for i in count(1))

        # Keep track of the objective variables that are non-zero
        self._non_zero_objective = set()

        self._result = None

    @property
    def cplex(self):
        """The underlying Cplex object"""
        return self._cp

    def define(self, *names, **kwargs):
        """Define variable in the problem.

        Variables must be defined before they can be accessed by var() or
        set(). This function takes keyword arguments lower and upper to define
        the bounds of the variable (default: -inf to inf). The keyword argument
        types can be used to select the type of the variable (Continuous
        (default), Binary or Integer). Setting any variables different than
        Continuous will turn the problem into an MILP problem.
        """
        names = tuple(names)
        lower = kwargs.get('lower', None)
        upper = kwargs.get('upper', None)
        vartype = kwargs.get('types', None)

        # Repeat values if a scalar is given
        if lower is None or isinstance(lower, numbers.Number):
            lower = repeat(lower, len(names))
        if upper is None or isinstance(upper, numbers.Number):
            upper = repeat(upper, len(names))
        if vartype is None or vartype in (
                VariableType.Continuous, VariableType.Binary,
                VariableType.Integer):
            vartype = repeat(vartype, len(names))

        lp_names = tuple(next(self._var_names) for name in names)

        # Assign default values
        lower = (-cp.infinity if value is None or value == -_INF
                 else value for value in lower)
        upper = (cp.infinity if value is None or value == _INF
                 else value for value in upper)
        vartype = tuple(VariableType.Continuous if value is None else value
                        for value in vartype)

        args = {'lb': tuple(lower), 'ub': tuple(upper)}
        if any(value != VariableType.Continuous for value in vartype):
            # Set types only if some are integer (otherwise Cplex will change
            # the solver to MILP).
            args['types'] = tuple(Problem.VARTYPE_MAP[t] for t in vartype)

        self._variables.update(zip(names, lp_names))
        self._cp.variables.add(**args)

    def has_variable(self, name):
        """Check whether variable is defined in the model."""
        return name in self._variables

    def _add_constraints(self, relation):
        """Add the given relation as one or more constraints

        Return a list of the names of the constraints added.
        """
        expression = relation.expression
        pairs = []
        for value_set in expression.value_sets():
            ind, val = zip(*((self._variables[variable], float(value))
                             for variable, value in value_set))
            pairs.append(cp.SparsePair(ind=ind, val=val))

        names = [next(self._constr_names) for _ in pairs]

        sense = self.CONSTR_SENSE_MAP[relation.sense]
        self._cp.linear_constraints.add(
            names=names, lin_expr=pairs,
            senses=tuple(repeat(sense, len(pairs))),
            rhs=tuple(repeat(float(-expression.offset), len(pairs))))

        return names

    def add_linear_constraints(self, *relations):
        """Add constraints to the problem

        Each constraint is represented by a Relation, and the
        expression in that relation can be a set expression.
        """
        constraints = []

        for relation in relations:
            if self._check_relation(relation):
                constraints.append(Constraint(self, None))
            else:
                for name in self._add_constraints(relation):
                    constraints.append(Constraint(self, name))

        return constraints

    def set_linear_objective(self, expression):
        """Set linear objective of problem"""

        if isinstance(expression, numbers.Number):
            # Allow expressions with no variables as objective,
            # represented as a number
            expression = Expression()

        # Reset previous objective. We have to build the set of variables to
        # update so that we can avoid calling set_linear if the set is empty.
        # This is due to set_linear failing if the input is an empty
        # iterable.
        reset_vars = set(
            self._variables[var] for var in self._non_zero_objective
            if var not in expression)
        if len(reset_vars) > 0:
            self._cp.objective.set_linear((var, 0) for var in reset_vars)

        # Set actual objective values
        if len(expression.values()) > 0:
            self._cp.objective.set_linear(
                (self._variables[var], value)
                for var, value in expression.values())

        # Keep track of new non-zeros
        self._non_zero_objective = set(expression.variables())

    def set_objective_sense(self, sense):
        """Set type of problem (maximize or minimize)"""
        if sense == ObjectiveSense.Minimize:
            self._cp.objective.set_sense(self._cp.objective.sense.minimize)
        elif sense == ObjectiveSense.Maximize:
            self._cp.objective.set_sense(self._cp.objective.sense.maximize)
        else:
            raise ValueError('Invalid objective sense')

    def _solve(self):
        self._cp.solve()
        if (self._cp.solution.get_status() ==
                self._cp.solution.status.abort_user):
            raise KeyboardInterrupt()

    def solve(self, sense=None):
        """Solve problem"""
        if sense is not None:
            self.set_objective_sense(sense)

        self._solve()
        self._result = Result(self)

        return self._result

    @property
    def result(self):
        return self._result


class Constraint(BaseConstraint):
    """Represents a constraint in a cplex.Problem"""

    def __init__(self, prob, name):
        self._prob = prob
        self._name = name

    def delete(self):
        if self._name is not None:
            self._prob._cp.linear_constraints.delete(self._name)


class Result(BaseResult):
    """Represents the solution to a cplex.Problem

    This object will be returned from the cplex.Problem.solve() method or by
    accessing the cplex.Problem.result property after solving a problem. This
    class should not be instantiated manually.

    Result will evaluate to a boolean according to the success of the
    solution, so checking the truth value of the result will immediately
    indicate whether solving was successful.
    """

    def __init__(self, prob):
        self._problem = prob

    def _check_valid(self):
        if self._problem.result != self:
            raise InvalidResultError()

    @property
    def success(self):
        """Return boolean indicating whether a solution was found"""
        self._check_valid()
        return self._problem._cp.solution.get_status() in (
            self._problem._cp.solution.status.optimal,
            self._problem._cp.solution.status.optimal_tolerance,
            self._problem._cp.solution.status.MIP_optimal)

    @property
    def status(self):
        """Return string indicating the error encountered on failure"""
        self._check_valid()
        return self._problem._cp.solution.get_status_string()

    @property
    def unbounded(self):
        """Whether solution is unbounded"""
        self._check_valid()

        cp = self._problem._cp
        status = cp.solution.get_status()
        presolve = cp.parameters.preprocessing.presolve.get()
        if (status == cp.solution.status.infeasible_or_unbounded and
                presolve):
            # Disable presolve to obtain a definitive answer
            logger.info('Disabling presolver and solving again to determine'
                        ' whether objective is unbounded.')
            cp.parameters.preprocessing.presolve.set(False)
            try:
                self._problem._solve()
            finally:
                cp.parameters.preprocessing.presolve.set(True)

            status = cp.solution.get_status()

        return status == cp.solution.status.unbounded

    def get_value(self, expression):
        """Return value of expression"""

        self._check_valid()
        if isinstance(expression, Expression):
            return sum(self._problem._cp.solution.get_values(
                self._problem._variables[var])*value
                for var, value in expression.values())
        elif expression not in self._problem._variables:
            raise ValueError('Unknown expression: {}'.format(expression))
        return self._problem._cp.solution.get_values(
            self._problem._variables[expression])



# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2014-2015  Jon Lund Steffensen <jon_steffensen@uri.edu>

import time
import logging

from ..command import Command, SolverCommandMixin, CommandError
from ..util import MaybeRelative
from .. import fluxanalysis

logger = logging.getLogger(__name__)


class FluxVariabilityCommand(SolverCommandMixin, Command):
    """Run flux variablity analysis on the model."""

    @classmethod
    def init_parser(cls, parser):
        parser.add_argument(
            '--threshold', help='Threshold of objective reaction flux',
            type=MaybeRelative, default=MaybeRelative('100%'))
        parser.add_argument(
            '--tfba', help='Enable thermodynamic constraints on FVA',
            action='store_true')
        parser.add_argument('reaction', help='Reaction to maximize', nargs='?')
        super(FluxVariabilityCommand, cls).init_parser(parser)

    def run(self):
        """Run flux variability command"""

        # Load compound information
        compound_name = {}
        for compound in self._model.parse_compounds():
            if 'name' in compound.properties:
                compound_name[compound.id] = compound.properties['name']
            elif compound.id not in compound_name:
                compound_name[compound.id] = compound.id

        if self._args.reaction is not None:
            reaction = self._args.reaction
        else:
            reaction = self._model.get_biomass_reaction()
            if reaction is None:
                raise CommandError('The biomass reaction was not specified')

        if not self._mm.has_reaction(reaction):
            raise CommandError('Specified reaction is not in model: {}'.format(
                reaction))

        enable_tfba = self._args.tfba
        if enable_tfba:
            solver = self._get_solver(integer=True)
        else:
            solver = self._get_solver()

        start_time = time.time()

        threshold = self._args.threshold
        if threshold.relative:
            fba_fluxes = dict(fluxanalysis.flux_balance(
                self._mm, reaction, tfba=False, solver=solver))
            threshold.reference = fba_fluxes[reaction]

        logger.info('Setting objective threshold to {}'.format(
            threshold))

        flux_bounds = fluxanalysis.flux_variability(
            self._mm, sorted(self._mm.reactions), {reaction: float(threshold)},
            tfba=enable_tfba, solver=solver)
        for reaction_id, bounds in flux_bounds:
            rx = self._mm.get_reaction(reaction_id)
            rxt = rx.translated_compounds(lambda x: compound_name.get(x, x))
            print('{}\t{}\t{}\t{}'.format(
                reaction_id, bounds[0], bounds[1], rxt))

        logger.info('Solving took {:.2f} seconds'.format(
            time.time() - start_time))



# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2015  Keith Dufault-Thompson <keitht547@my.uri.edu>

import logging

from ..command import Command, CommandError

try:
    import xlsxwriter
except ImportError:
    xlsxwriter = None

logger = logging.getLogger(__name__)


class ExcelExportCommand(Command):
    """Export the metabolic model as an Excel workbook"""

    @classmethod
    def init_parser(cls, parser):
        parser.add_argument(
            'file', type=str, help='File path for writing the Excel workbook')

    def run(self):
        model = self._model
        if xlsxwriter is None:
            raise CommandError(
                'Excel export requires the XlsxWriter python module')
        workbook = xlsxwriter.Workbook(self._args.file)
        reaction_sheet = workbook.add_worksheet(name='Reactions')

        property_set = set()
        for j in model.parse_reactions():
            property_set.update(j.properties)
        property_list = list(property_set)
        property_list_sorted = sorted(property_list,
                                      key=lambda x: (x != 'id',
                                                     x != 'equation', x))
        for z, i in enumerate(property_list_sorted):
            reaction_sheet.write_string(0, z, str(i))
        for x, i in enumerate(model.parse_reactions()):
            for y, j in enumerate(property_list_sorted):
                reaction_sheet.write_string(x+1, y, str(i.properties.get(j)))

        compound_sheet = workbook.add_worksheet(name='Compounds')

        compound_set = set()
        for compound in model.parse_compounds():
            compound_set.update(compound.properties)

        compound_list_sorted = sorted(compound_set,
                                      key=lambda x: (x != 'id',
                                                     x != 'name', x))

        for z, i in enumerate(compound_list_sorted):
            compound_sheet.write_string(0, z, str(i))
        for x, i in enumerate(model.parse_compounds()):
            for y, j in enumerate(compound_list_sorted):
                compound_sheet.write_string(x+1, y, str(i.properties.get(j)))

        media_sheet = workbook.add_worksheet(name='Medium')

        media_sheet.write_string(0, 0, 'Compound ID')
        media_sheet.write_string(0, 1, 'Reaction ID')
        media_sheet.write_string(0, 2, 'Lower Limit')
        media_sheet.write_string(0, 3, 'Upper Limit')

        default_flux = model.get_default_flux_limit()
        if default_flux is None:
            default_flux = 1000

        for x, (compound, reaction, lower, upper) in enumerate(
                model.parse_medium()):
            if lower is None:
                lower = -1 * default_flux

            if upper is None:
                upper = default_flux

            media_sheet.write(x+1, 0, str(compound))
            media_sheet.write(x+1, 1, str(reaction))
            media_sheet.write(x+1, 2, str(lower))
            media_sheet.write(x+1, 3, str(upper))

        limits_sheet = workbook.add_worksheet(name='Limits')

        limits_sheet.write_string(0, 0, 'Reaction ID')
        limits_sheet.write_string(0, 1, 'Lower Limit')
        limits_sheet.write_string(0, 2, str('Upper Limit'))

        for x, i in enumerate(model.parse_limits()):
            limits_sheet.write(x, 0, (i[0]))
            limits_sheet.write(x, 1, (i[1]))
            limits_sheet.write(x, 2, (i[2]))

        model_info = workbook.add_worksheet(name='Model Info')

        model_info.write(0, 0, ('Model Name: {}'.
                                format(model.get_name())))
        model_info.write(1, 0, ('Biomass Reaction: {}'.
                                format(model.get_biomass_reaction())))
        model_info.write(2, 0, ('Default Flux Limits: {}'.
                                format(model.get_default_flux_limit())))

        workbook.close()



# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2014-2015  Jon Lund Steffensen <jon_steffensen@uri.edu>

"""Various utilities."""

import re
import math
import subprocess
import collections

from six import iteritems


class LoggerFile(object):
    """File-like object that forwards to a logger.

    The Cplex API takes a file-like object for writing log output.
    This class allows us to forward the Cplex messages to the
    Python logging system.
    """

    def __init__(self, logger, level):
        self._logger = logger
        self._level = level

    def write(self, s):
        """Write message to logger."""
        for line in re.split(r'\n+', s):
            if line != '':
                self._logger.log(self._level, line)

    def flush(self):
        """Flush stream.

        This is a noop."""


class MaybeRelative(object):
    """Helper type for parsing possibly relative parameters.

    >>> arg = MaybeRelative('40%')
    >>> arg.reference = 200.0
    >>> float(arg)
    80.0

    >>> arg = MaybeRelative('24.5')
    >>> arg.reference = 150.0
    >>> float(arg)
    24.5
    """

    def __init__(self, s):
        try:
            self._value = float(s)
            self._relative = False
        except ValueError:
            self._value = self._parse_percentage(s)
            self._relative = self._value != 0.0

        self._reference = None

    @classmethod
    def _parse_percentage(cls, s):
        """Parse string as a percentage (e.g. '42.0%') and return as float."""
        m = re.match('^(.+)%$', s)
        if not m:
            raise ValueError('Unable to parse as percentage: {}'.format(
                repr(s)))

        return float(m.group(1)) / 100.0

    @property
    def relative(self):
        """Whether the parsed number was relative."""
        return self._relative

    @property
    def reference(self):
        """The reference used for converting to absolute value."""
        return self._reference

    @reference.setter
    def reference(self, value):
        self._reference = value

    def __float__(self):
        if self._relative:
            if self._reference is None:
                raise ValueError('Reference not set!')
            return self._reference * self._value
        else:
            return self._value

    def __str__(self):
        if self._relative:
            f = None if self._reference is None else float(self)
            return '{:.1%} of {} = {}'.format(
                self._value, self._reference, f)
        else:
            return str(self._value)

    def __repr__(self):
        return '<{}, {}>'.format(self.__class__.__name__, str(self))


class FrozenOrderedSet(collections.Set, collections.Hashable):
    """An immutable set that retains insertion order."""

    def __init__(self, seq=[]):
        self.__map = collections.OrderedDict()
        for e in seq:
            self.__map[e] = None

    def __contains__(self, element):
        return element in self.__map

    def __iter__(self):
        return iter(self.__map)

    def __len__(self):
        return len(self.__map)

    def __hash__(self):
        h = 0
        for e in self:
            h ^= 31 * hash(e)
        return h


def git_try_describe(repo_path):
    """Try to describe the current commit of a Git repository.

    Return a string containing a string with the commit ID and/or a base tag,
    if successful. Otherwise, return None.
    """
    try:
        p = subprocess.Popen(['git', 'describe', '--always', '--dirty'],
                             cwd=repo_path, stdout=subprocess.PIPE,
                             stderr=subprocess.PIPE)
        output, _ = p.communicate()
    except:
        return None
    else:
        if p.returncode == 0:
            return output.strip()

    return None


def convex_cardinality_relaxed(f, epsilon=1e-5):
    """Transform L1-norm optimization function into cardinality optimization.

    The given function must optimize a convex problem with
    a weighted L1-norm as the objective. The transformed function
    will apply the iterated weighted L1 heuristic to approximately
    optimize the cardinality of the solution. This method is
    described by S. Boyd, "L1-norm norm methods for convex cardinality
    problems." Lecture Notes for EE364b, Stanford University, 2007.
    Available online at www.stanford.edu/class/ee364b/.

    The given function must take an optional keyword parameter weights
    (dictionary), and the weights must be set to one if not specified.
    The function must return the non-weighted solution as an iterator
    over (identifier, value)-tuples, either directly or as the first
    element of a tuple.
    """

    def convex_cardinality_wrapper(*args, **kwargs):
        def dict_result(r):
            if isinstance(r, tuple):
                return dict(r[0])
            return dict(r)

        # Initial run with default weights
        full_result = f(*args, **kwargs)
        result = dict_result(full_result)

        def update_weight(value):
            return 1/(epsilon + abs(value))

        # Iterate until the difference from one iteration to
        # the next is less than epsilon.
        while True:
            weights = {identifier: update_weight(value)
                       for identifier, value in iteritems(result)}
            kwargs['weights'] = weights

            last_result = result
            full_result = f(*args, **kwargs)
            result = dict_result(full_result)

            delta = math.sqrt(sum(pow(value - last_result[identifier], 2)
                                  for identifier, value in iteritems(result)))
            if delta < epsilon:
                break

        if isinstance(full_result, tuple):
            return (iteritems(result),) + full_result[1:]
        return iteritems(result)

    return convex_cardinality_wrapper



#!/usr/bin/env python
# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2014-2015  Jon Lund Steffensen <jon_steffensen@uri.edu>

import unittest

from psamm.datasource import sbml
from psamm.reaction import Reaction, Compound

from decimal import Decimal
from fractions import Fraction
from six import StringIO


class TestSBMLDatabaseL1V2(unittest.TestCase):
    """Test parsing of a simple level 1 version 2 SBML file"""

    def setUp(self):
        s = StringIO('''<?xml version="1.0" encoding="UTF-8"?>
<sbml xmlns="http://www.sbml.org/sbml/level1"
      xmlns:html="http://www.w3.org/1999/xhtml"
      level="1" version="2">
 <model name="Test model">
  <listOfCompartments>
   <compartment name="cell"/>
  </listOfCompartments>
  <listOfSpecies>
   <species name="Glucose" compartment="cell" initialAmount="1"/>
   <species name="Glucose_6_P" compartment="cell" initialAmount="1"/>
   <species name="H2O" compartment="cell" initialAmount="1"/>
   <species name="Phosphate" compartment="cell" initialAmount="1" boundaryCondition="false"/>
   <species name="Biomass" compartment="cell" initialAmount="1" boundaryCondition="true"/>
  </listOfSpecies>
  <listOfReactions>
   <reaction name="G6Pase" reversible="true">
    <listOfReactants>
     <speciesReference species="Glucose" stoichiometry="2"/>
     <speciesReference species="Phosphate" stoichiometry="2"/>
    </listOfReactants>
    <listOfProducts>
     <speciesReference species="H2O" stoichiometry="2"/>
     <speciesReference species="Glucose_6_P" stoichiometry="2"/>
    </listOfProducts>
    <notes>
     <html:p>Glucose 6-phosphatase</html:p>
    </notes>
   </reaction>
   <reaction name="Biomass" reversible="false">
    <listOfReactants>
     <speciesReference species="Glucose_6_P" stoichiometry="56" denominator="100"/>
    </listOfReactants>
    <listOfProducts>
     <speciesReference species="Biomass"/>
    </listOfProducts>
   </reaction>
  </listOfReactions>
 </model>
</sbml>''')
        self.reader = sbml.SBMLReader(s)

    def test_model_name(self):
        self.assertEqual(self.reader.name, 'Test model')

    def test_compounds_exist(self):
        species = {entry.id: entry for entry in self.reader.species}
        self.assertEqual(len(species), 5)

        self.assertEqual(species['Glucose'].id, 'Glucose')
        self.assertEqual(species['Glucose'].name, 'Glucose')
        self.assertEqual(species['Glucose'].compartment, 'cell')
        self.assertFalse(species['Glucose'].boundary)

        self.assertEqual(species['Glucose_6_P'].id, 'Glucose_6_P')
        self.assertEqual(species['Glucose_6_P'].name, 'Glucose_6_P')
        self.assertEqual(species['Glucose_6_P'].compartment, 'cell')
        self.assertFalse(species['Glucose_6_P'].boundary)

        self.assertEqual(species['H2O'].id, 'H2O')
        self.assertEqual(species['H2O'].name, 'H2O')
        self.assertEqual(species['H2O'].compartment, 'cell')
        self.assertFalse(species['H2O'].boundary)

        self.assertFalse(species['Phosphate'].boundary)
        self.assertTrue(species['Biomass'].boundary)

    def test_g6pase_reaction_exists(self):
        reaction = self.reader.get_reaction('G6Pase')
        self.assertTrue(reaction.reversible)

        # Compare equation of reaction
        actual_equation = Reaction(Reaction.Bidir,
                                   [(Compound('Glucose', 'cell'), 2),
                                    (Compound('Phosphate', 'cell'), 2)],
                                   [(Compound('H2O', 'cell'), 2),
                                    (Compound('Glucose_6_P', 'cell'), 2)])
        self.assertEqual(reaction.equation, actual_equation)

    def test_biomass_reaction_exists(self):
        reaction = self.reader.get_reaction('Biomass')
        self.assertFalse(reaction.reversible)

        # Compare equation of reaction
        actual_equation = Reaction(Reaction.Right,
                                   [(Compound('Glucose_6_P', 'cell'),
                                     Fraction(56, 100))],
                                   [(Compound('Biomass', 'cell'), 1)])
        self.assertEqual(reaction.equation, actual_equation)

    def test_reaction_xml_notes(self):
        reaction = self.reader.get_reaction('G6Pase')
        notes = reaction.xml_notes

        notes_tags = list(notes)
        self.assertEqual(len(notes_tags), 1)
        self.assertEqual(notes_tags[0].tag, '{http://www.w3.org/1999/xhtml}p')
        self.assertEqual(notes_tags[0].text, 'Glucose 6-phosphatase')

    def test_objective_not_present(self):
        objectives = list(self.reader.objectives)
        self.assertEqual(len(objectives), 0)
        self.assertIsNone(self.reader.get_active_objective())

    def test_flux_bounds_not_present(self):
        flux_bounds = list(self.reader.flux_bounds)
        self.assertEqual(len(flux_bounds), 0)


class TestSBMLDatabaseL2V5(unittest.TestCase):
    """Test parsing of a simple level 2 version 5 SBML file"""

    def setUp(self):
        s = StringIO('''<?xml version="1.0" encoding="UTF-8"?>
<sbml xmlns="http://www.sbml.org/sbml/level2/version5"
      xmlns:html="http://www.w3.org/1999/xhtml"
      level="2" version="5">
 <model id="test_model" name="Test model">
  <listOfCompartments>
   <compartment id="C_c" name="cell"/>
  </listOfCompartments>
  <listOfSpecies>
   <species id="M_Glucose" name="Glucose" compartment="C_c"/>
   <species id="M_Glucose_6_P" name="Glucose-6-P" compartment="C_c"/>
   <species id="M_H2O" name="H2O" compartment="C_c"/>
   <species id="M_Phosphate" name="Phosphate" compartment="C_c" boundaryCondition="false"/>
   <species id="M_Biomass" name="Biomass" compartment="C_c" boundaryCondition="true"/>
  </listOfSpecies>
  <listOfReactions>
   <reaction id="R_G6Pase" reversible="true">
    <listOfReactants>
     <speciesReference species="M_Glucose" stoichiometry="2"/>
     <speciesReference species="M_Phosphate" stoichiometry="2"/>
    </listOfReactants>
    <listOfProducts>
     <speciesReference species="M_H2O" stoichiometry="2"/>
     <speciesReference species="M_Glucose_6_P" stoichiometry="2"/>
    </listOfProducts>
    <notes>
     <html:p>Glucose 6-phosphatase</html:p>
    </notes>
   </reaction>
   <reaction id="R_Biomass" reversible="false">
    <listOfReactants>
     <speciesReference species="M_Glucose_6_P" stoichiometry="0.56"/>
    </listOfReactants>
    <listOfProducts>
     <speciesReference species="M_Biomass"/>
    </listOfProducts>
   </reaction>
  </listOfReactions>
 </model>
</sbml>''')
        self.reader = sbml.SBMLReader(s)

    def test_model_name(self):
        self.assertEqual(self.reader.id, 'test_model')
        self.assertEqual(self.reader.name, 'Test model')

    def test_compounds_exist(self):
        species = {entry.id: entry for entry in self.reader.species}
        self.assertEqual(len(species), 5)

        self.assertEqual(species['M_Glucose'].id, 'M_Glucose')
        self.assertEqual(species['M_Glucose'].name, 'Glucose')
        self.assertEqual(species['M_Glucose'].compartment, 'C_c')
        self.assertFalse(species['M_Glucose'].boundary)

        self.assertEqual(species['M_Glucose_6_P'].id, 'M_Glucose_6_P')
        self.assertEqual(species['M_Glucose_6_P'].name, 'Glucose-6-P')
        self.assertEqual(species['M_Glucose_6_P'].compartment, 'C_c')
        self.assertFalse(species['M_Glucose_6_P'].boundary)

        self.assertEqual(species['M_H2O'].id, 'M_H2O')
        self.assertEqual(species['M_H2O'].name, 'H2O')
        self.assertEqual(species['M_H2O'].compartment, 'C_c')
        self.assertFalse(species['M_H2O'].boundary)

        self.assertFalse(species['M_Phosphate'].boundary)
        self.assertTrue(species['M_Biomass'].boundary)

    def test_g6pase_reaction_exists(self):
        reaction = self.reader.get_reaction('R_G6Pase')
        self.assertTrue(reaction.reversible)

        # Compare equation of reaction
        actual_equation = Reaction(Reaction.Bidir,
                                   [(Compound('M_Glucose', 'C_c'), 2),
                                    (Compound('M_Phosphate', 'C_c'), 2)],
                                   [(Compound('M_H2O', 'C_c'), 2),
                                    (Compound('M_Glucose_6_P', 'C_c'), 2)])
        self.assertEqual(reaction.equation, actual_equation)

    def test_biomass_reaction_exists(self):
        reaction = self.reader.get_reaction('R_Biomass')
        self.assertFalse(reaction.reversible)

        # Compare equation of reaction
        actual_equation = Reaction(Reaction.Right,
                                   [(Compound('M_Glucose_6_P', 'C_c'),
                                     Decimal('0.56'))],
                                   [(Compound('M_Biomass', 'C_c'), 1)])
        self.assertEqual(reaction.equation, actual_equation)

    def test_reaction_xml_notes(self):
        reaction = self.reader.get_reaction('R_G6Pase')
        notes = reaction.xml_notes

        notes_tags = list(notes)
        self.assertEqual(len(notes_tags), 1)
        self.assertEqual(notes_tags[0].tag, '{http://www.w3.org/1999/xhtml}p')
        self.assertEqual(notes_tags[0].text, 'Glucose 6-phosphatase')

    def test_objective_not_present(self):
        objectives = list(self.reader.objectives)
        self.assertEqual(len(objectives), 0)
        self.assertIsNone(self.reader.get_active_objective())

    def test_flux_bounds_not_present(self):
        flux_bounds = list(self.reader.flux_bounds)
        self.assertEqual(len(flux_bounds), 0)


class TestSBMLDatabaseL3V1(unittest.TestCase):
    """Test parsing of a simple level 3 version 1 SBML file"""

    def setUp(self):
        s = StringIO('''<?xml version="1.0" encoding="UTF-8"?>
<sbml xmlns="http://www.sbml.org/sbml/level3/version1/core"
      xmlns:html="http://www.w3.org/1999/xhtml"
      level="3" version="1">
 <model id="test_model" name="Test model">
  <listOfCompartments>
   <compartment id="C_c" name="cell" constant="true"/>
  </listOfCompartments>
  <listOfSpecies>
   <species id="M_Glucose" name="Glucose" compartment="C_c" constant="false" boundaryCondition="false" hasOnlySubstanceUnits="false"/>
   <species id="M_Glucose_6_P" name="Glucose-6-P" compartment="C_c" constant="false" boundaryCondition="false" hasOnlySubstanceUnits="false"/>
   <species id="M_H2O" name="H2O" compartment="C_c" constant="false" boundaryCondition="false" hasOnlySubstanceUnits="false"/>
   <species id="M_Phosphate" name="Phosphate" compartment="C_c" constant="false" boundaryCondition="false" hasOnlySubstanceUnits="false"/>
   <species id="M_Biomass" name="Biomass" compartment="C_c" constant="false" boundaryCondition="true" hasOnlySubstanceUnits="false"/>
  </listOfSpecies>
  <listOfReactions>
   <reaction id="R_G6Pase" reversible="true" fast="false">
    <listOfReactants>
     <speciesReference species="M_Glucose" stoichiometry="2" constant="true"/>
     <speciesReference species="M_Phosphate" stoichiometry="2" constant="true"/>
    </listOfReactants>
    <listOfProducts>
     <speciesReference species="M_H2O" stoichiometry="2" constant="true"/>
     <speciesReference species="M_Glucose_6_P" stoichiometry="2" constant="true"/>
    </listOfProducts>
    <notes>
     <html:p>Glucose 6-phosphatase</html:p>
    </notes>
   </reaction>
   <reaction id="R_Biomass" reversible="false" fast="false">
    <listOfReactants>
     <speciesReference species="M_Glucose_6_P" stoichiometry="0.56" constant="true"/>
    </listOfReactants>
    <listOfProducts>
     <speciesReference species="M_Biomass" stoichiometry="1" constant="true"/>
    </listOfProducts>
   </reaction>
  </listOfReactions>
 </model>
</sbml>''')
        self.reader = sbml.SBMLReader(s)

    def test_model_name(self):
        self.assertEqual(self.reader.id, 'test_model')
        self.assertEqual(self.reader.name, 'Test model')

    def test_compounds_exist(self):
        species = {entry.id: entry for entry in self.reader.species}
        self.assertEqual(len(species), 5)

        self.assertEqual(species['M_Glucose'].id, 'M_Glucose')
        self.assertEqual(species['M_Glucose'].name, 'Glucose')
        self.assertEqual(species['M_Glucose'].compartment, 'C_c')
        self.assertFalse(species['M_Glucose'].boundary)

        self.assertEqual(species['M_Glucose_6_P'].id, 'M_Glucose_6_P')
        self.assertEqual(species['M_Glucose_6_P'].name, 'Glucose-6-P')
        self.assertEqual(species['M_Glucose_6_P'].compartment, 'C_c')
        self.assertFalse(species['M_Glucose_6_P'].boundary)

        self.assertEqual(species['M_H2O'].id, 'M_H2O')
        self.assertEqual(species['M_H2O'].name, 'H2O')
        self.assertEqual(species['M_H2O'].compartment, 'C_c')
        self.assertFalse(species['M_H2O'].boundary)

        self.assertFalse(species['M_Phosphate'].boundary)
        self.assertTrue(species['M_Biomass'].boundary)

    def test_g6pase_reaction_exists(self):
        reaction = self.reader.get_reaction('R_G6Pase')
        self.assertTrue(reaction.reversible)

        # Compare equation of reaction
        actual_equation = Reaction(Reaction.Bidir,
                                   [(Compound('M_Glucose', 'C_c'), 2),
                                    (Compound('M_Phosphate', 'C_c'), 2)],
                                   [(Compound('M_H2O', 'C_c'), 2),
                                    (Compound('M_Glucose_6_P', 'C_c'), 2)])
        self.assertEqual(reaction.equation, actual_equation)

    def test_biomass_reaction_exists(self):
        reaction = self.reader.get_reaction('R_Biomass')
        self.assertFalse(reaction.reversible)

        # Compare equation of reaction
        actual_equation = Reaction(Reaction.Right,
                                   [(Compound('M_Glucose_6_P', 'C_c'),
                                     Decimal('0.56'))],
                                   [(Compound('M_Biomass', 'C_c'), 1)])
        self.assertEqual(reaction.equation, actual_equation)

    def test_reaction_xml_notes(self):
        reaction = self.reader.get_reaction('R_G6Pase')
        notes = reaction.xml_notes

        notes_tags = list(notes)
        self.assertEqual(len(notes_tags), 1)
        self.assertEqual(notes_tags[0].tag, '{http://www.w3.org/1999/xhtml}p')
        self.assertEqual(notes_tags[0].text, 'Glucose 6-phosphatase')

    def test_objective_not_present(self):
        objectives = list(self.reader.objectives)
        self.assertEqual(len(objectives), 0)
        self.assertIsNone(self.reader.get_active_objective())

    def test_flux_bounds_not_present(self):
        flux_bounds = list(self.reader.flux_bounds)
        self.assertEqual(len(flux_bounds), 0)


class TestSBMLDatabaseL3V1WithFBCV1(unittest.TestCase):
    """Test parsing of a level 3 version 1 SBML file with FBC version 1"""

    def setUp(self):
        s = StringIO('''<?xml version="1.0" encoding="UTF-8"?>
<sbml xmlns="http://www.sbml.org/sbml/level3/version1/core"
      xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version1"
      xmlns:html="http://www.w3.org/1999/xhtml"
      level="3" version="1"
      fbc:required="false">
 <model id="test_model" name="Test model">
  <listOfCompartments>
   <compartment id="C_c" name="cell" constant="true"/>
  </listOfCompartments>
  <listOfSpecies>
   <species id="M_Glucose" name="Glucose" compartment="C_c" constant="false" boundaryCondition="false" hasOnlySubstanceUnits="false" fbc:charge="0" fbc:chemicalFormula="C6H12O6"/>
   <species id="M_Glucose_6_P" name="Glucose-6-P" compartment="C_c" constant="false" boundaryCondition="false" hasOnlySubstanceUnits="false" fbc:charge="-2" fbc:chemicalFormula="C6H11O9P"/>
   <species id="M_H2O" name="H2O" compartment="C_c" constant="false" boundaryCondition="false" hasOnlySubstanceUnits="false" fbc:charge="0" fbc:chemicalFormula="H2O"/>
   <species id="M_Phosphate" name="Phosphate" compartment="C_c" constant="false" boundaryCondition="false" hasOnlySubstanceUnits="false" fbc:charge="-2" fbc:chemicalFormula="HO4P"/>
   <species id="M_Biomass" name="Biomass" compartment="C_c" constant="false" boundaryCondition="true" hasOnlySubstanceUnits="false"/>
  </listOfSpecies>
  <listOfReactions>
   <reaction id="R_G6Pase" reversible="true" fast="false">
    <listOfReactants>
     <speciesReference species="M_Glucose" stoichiometry="2" constant="true"/>
     <speciesReference species="M_Phosphate" stoichiometry="2" constant="true"/>
    </listOfReactants>
    <listOfProducts>
     <speciesReference species="M_H2O" stoichiometry="2" constant="true"/>
     <speciesReference species="M_Glucose_6_P" stoichiometry="2" constant="true"/>
    </listOfProducts>
   </reaction>
   <reaction id="R_Biomass" reversible="false" fast="false">
    <listOfReactants>
     <speciesReference species="M_Glucose_6_P" stoichiometry="0.56" constant="true"/>
    </listOfReactants>
    <listOfProducts>
     <speciesReference species="M_Biomass" stoichiometry="1" constant="true"/>
    </listOfProducts>
   </reaction>
  </listOfReactions>
  <fbc:listOfObjectives fbc:activeObjective="obj1">
   <fbc:objective fbc:id="obj1" fbc:name="Objective 1" fbc:type="maximize">
    <fbc:listOfFluxObjectives>
     <fbc:fluxObjective fbc:reaction="R_Biomass" fbc:coefficient="1"/>
    </fbc:listOfFluxObjectives>
   </fbc:objective>
  </fbc:listOfObjectives>
  <fbc:listOfFluxBounds>
   <fbc:fluxBound fbc:reaction="R_G6Pase" fbc:operation="greaterEqual" fbc:value="-10"/>
   <fbc:fluxBound fbc:reaction="R_G6Pase" fbc:operation="lessEqual" fbc:value="1000"/>
   <fbc:fluxBound fbc:reaction="R_Biomass" fbc:operation="greaterEqual" fbc:value="0"/>
   <fbc:fluxBound fbc:reaction="R_Biomass" fbc:operation="lessEqual" fbc:value="1000"/>
  </fbc:listOfFluxBounds>
 </model>
</sbml>''')
        self.reader = sbml.SBMLReader(s)

    def test_model_name(self):
        self.assertEqual(self.reader.id, 'test_model')
        self.assertEqual(self.reader.name, 'Test model')

    def test_compounds_exist(self):
        species = {entry.id: entry for entry in self.reader.species}
        self.assertEqual(len(species), 5)

        self.assertEqual(species['M_Glucose'].id, 'M_Glucose')
        self.assertEqual(species['M_Glucose'].name, 'Glucose')
        self.assertEqual(species['M_Glucose'].compartment, 'C_c')
        self.assertFalse(species['M_Glucose'].boundary)
        self.assertEqual(species['M_Glucose'].formula, 'C6H12O6')
        self.assertEqual(species['M_Glucose'].charge, 0)

        self.assertEqual(species['M_Glucose_6_P'].id, 'M_Glucose_6_P')
        self.assertEqual(species['M_Glucose_6_P'].name, 'Glucose-6-P')
        self.assertEqual(species['M_Glucose_6_P'].compartment, 'C_c')
        self.assertFalse(species['M_Glucose_6_P'].boundary)
        self.assertEqual(species['M_Glucose_6_P'].formula, 'C6H11O9P')
        self.assertEqual(species['M_Glucose_6_P'].charge, -2)

        self.assertEqual(species['M_H2O'].id, 'M_H2O')
        self.assertEqual(species['M_H2O'].name, 'H2O')
        self.assertEqual(species['M_H2O'].compartment, 'C_c')
        self.assertFalse(species['M_H2O'].boundary)
        self.assertEqual(species['M_H2O'].formula, 'H2O')
        self.assertEqual(species['M_H2O'].charge, 0)

        self.assertFalse(species['M_Phosphate'].boundary)
        self.assertTrue(species['M_Biomass'].boundary)

        self.assertIsNone(species['M_Biomass'].formula)
        self.assertIsNone(species['M_Biomass'].charge)

    def test_g6pase_reaction_exists(self):
        reaction = self.reader.get_reaction('R_G6Pase')
        self.assertTrue(reaction.reversible)

        # Compare equation of reaction
        actual_equation = Reaction(Reaction.Bidir,
                                   [(Compound('M_Glucose', 'C_c'), 2),
                                    (Compound('M_Phosphate', 'C_c'), 2)],
                                   [(Compound('M_H2O', 'C_c'), 2),
                                    (Compound('M_Glucose_6_P', 'C_c'), 2)])
        self.assertEqual(reaction.equation, actual_equation)

        self.assertEqual(reaction.properties['lower_flux'], -10)
        self.assertEqual(reaction.properties['upper_flux'], 1000)

    def test_biomass_reaction_exists(self):
        reaction = self.reader.get_reaction('R_Biomass')
        self.assertFalse(reaction.reversible)

        # Compare equation of reaction
        actual_equation = Reaction(Reaction.Right,
                                   [(Compound('M_Glucose_6_P', 'C_c'),
                                     Decimal('0.56'))],
                                   [(Compound('M_Biomass', 'C_c'), 1)])
        self.assertEqual(reaction.equation, actual_equation)

        self.assertEqual(reaction.properties['lower_flux'], 0)
        self.assertEqual(reaction.properties['upper_flux'], 1000)

    def test_objective_exists(self):
        objectives = {entry.id: entry for entry in self.reader.objectives}
        self.assertEqual(len(objectives), 1)

        objective = objectives['obj1']
        self.assertEqual(objective.name, 'Objective 1')
        self.assertEqual(objective.type, 'maximize')
        self.assertEqual(dict(objective.reactions), {'R_Biomass': 1})

    def test_active_objective(self):
        objectives = {entry.id: entry for entry in self.reader.objectives}
        self.assertEqual(self.reader.get_active_objective(),
                         objectives['obj1'])

    def test_flux_bounds_exists(self):
        flux_bounds = list(self.reader.flux_bounds)
        self.assertEqual(len(flux_bounds), 4)

        biomass_bounds = set(
            (b.operation, b.value) for b in flux_bounds
            if b.reaction == 'R_Biomass')
        self.assertEqual(biomass_bounds, {
            ('greaterEqual', 0),
            ('lessEqual', 1000)})

        g6pase_bounds = set(
            (b.operation, b.value) for b in flux_bounds
            if b.reaction == 'R_G6Pase')
        self.assertEqual(g6pase_bounds, {
            ('greaterEqual', -10),
            ('lessEqual', 1000)})


class TestSBMLDatabaseL3V1WithFBCV2(unittest.TestCase):
    """Test parsing of a level 3 version 1 SBML file with FBC version 2"""

    def setUp(self):
        s = StringIO('''<?xml version="1.0" encoding="UTF-8"?>
<sbml xmlns="http://www.sbml.org/sbml/level3/version1/core"
      xmlns:fbc="http://www.sbml.org/sbml/level3/version1/fbc/version2"
      xmlns:html="http://www.w3.org/1999/xhtml"
      level="3" version="1"
      fbc:required="false">
 <model id="test_model" name="Test model">
  <listOfCompartments>
   <compartment id="C_c" name="cell" constant="true"/>
  </listOfCompartments>
  <listOfSpecies>
   <species id="M_Glucose" name="Glucose" compartment="C_c" constant="false" boundaryCondition="false" hasOnlySubstanceUnits="false" fbc:charge="0" fbc:chemicalFormula="C6H12O6"/>
   <species id="M_Glucose_6_P" name="Glucose-6-P" compartment="C_c" constant="false" boundaryCondition="false" hasOnlySubstanceUnits="false" fbc:charge="-2" fbc:chemicalFormula="C6H11O9P"/>
   <species id="M_H2O" name="H2O" compartment="C_c" constant="false" boundaryCondition="false" hasOnlySubstanceUnits="false" fbc:charge="0" fbc:chemicalFormula="H2O"/>
   <species id="M_Phosphate" name="Phosphate" compartment="C_c" constant="false" boundaryCondition="false" hasOnlySubstanceUnits="false" fbc:charge="-2" fbc:chemicalFormula="HO4P"/>
   <species id="M_Biomass" name="Biomass" compartment="C_c" constant="false" boundaryCondition="true" hasOnlySubstanceUnits="false"/>
  </listOfSpecies>
  <listOfParameters>
   <parameter constant="true" id="P_lower_G6Pase" value="-10"/>
   <parameter constant="true" id="P_lower_zero" value="0"/>
   <parameter constant="true" id="P_upper_default" value="1000"/>
  </listOfParameters>
  <listOfReactions>
   <reaction id="R_G6Pase" reversible="true" fast="false" fbc:lowerFluxBound="P_lower_G6Pase" fbc:upperFluxBound="P_upper_default">
    <listOfReactants>
     <speciesReference species="M_Glucose" stoichiometry="2" constant="true"/>
     <speciesReference species="M_Phosphate" stoichiometry="2" constant="true"/>
    </listOfReactants>
    <listOfProducts>
     <speciesReference species="M_H2O" stoichiometry="2" constant="true"/>
     <speciesReference species="M_Glucose_6_P" stoichiometry="2" constant="true"/>
    </listOfProducts>
   </reaction>
   <reaction id="R_Biomass" reversible="false" fast="false" fbc:lowerFluxBound="P_lower_zero" fbc:upperFluxBound="P_upper_default">
    <listOfReactants>
     <speciesReference species="M_Glucose_6_P" stoichiometry="0.56" constant="true"/>
    </listOfReactants>
    <listOfProducts>
     <speciesReference species="M_Biomass" stoichiometry="1" constant="true"/>
    </listOfProducts>
   </reaction>
  </listOfReactions>
  <fbc:listOfObjectives fbc:activeObjective="obj1">
   <fbc:objective fbc:id="obj1" fbc:name="Objective 1" fbc:type="maximize">
    <fbc:listOfFluxObjectives>
     <fbc:fluxObjective fbc:reaction="R_Biomass" fbc:coefficient="1"/>
    </fbc:listOfFluxObjectives>
   </fbc:objective>
  </fbc:listOfObjectives>
 </model>
</sbml>''')
        self.reader = sbml.SBMLReader(s)

    def test_model_name(self):
        self.assertEqual(self.reader.id, 'test_model')
        self.assertEqual(self.reader.name, 'Test model')

    def test_compounds_exist(self):
        species = {entry.id: entry for entry in self.reader.species}
        self.assertEqual(len(species), 5)

        self.assertEqual(species['M_Glucose'].id, 'M_Glucose')
        self.assertEqual(species['M_Glucose'].name, 'Glucose')
        self.assertEqual(species['M_Glucose'].compartment, 'C_c')
        self.assertFalse(species['M_Glucose'].boundary)
        self.assertEqual(species['M_Glucose'].formula, 'C6H12O6')
        self.assertEqual(species['M_Glucose'].charge, 0)

        self.assertEqual(species['M_Glucose_6_P'].id, 'M_Glucose_6_P')
        self.assertEqual(species['M_Glucose_6_P'].name, 'Glucose-6-P')
        self.assertEqual(species['M_Glucose_6_P'].compartment, 'C_c')
        self.assertFalse(species['M_Glucose_6_P'].boundary)
        self.assertEqual(species['M_Glucose_6_P'].formula, 'C6H11O9P')
        self.assertEqual(species['M_Glucose_6_P'].charge, -2)

        self.assertEqual(species['M_H2O'].id, 'M_H2O')
        self.assertEqual(species['M_H2O'].name, 'H2O')
        self.assertEqual(species['M_H2O'].compartment, 'C_c')
        self.assertFalse(species['M_H2O'].boundary)
        self.assertEqual(species['M_H2O'].formula, 'H2O')
        self.assertEqual(species['M_H2O'].charge, 0)

        self.assertFalse(species['M_Phosphate'].boundary)
        self.assertTrue(species['M_Biomass'].boundary)

        self.assertIsNone(species['M_Biomass'].formula)
        self.assertIsNone(species['M_Biomass'].charge)

    def test_g6pase_reaction_exists(self):
        reaction = self.reader.get_reaction('R_G6Pase')
        self.assertTrue(reaction.reversible)

        # Compare equation of reaction
        actual_equation = Reaction(Reaction.Bidir,
                                   [(Compound('M_Glucose', 'C_c'), 2),
                                    (Compound('M_Phosphate', 'C_c'), 2)],
                                   [(Compound('M_H2O', 'C_c'), 2),
                                    (Compound('M_Glucose_6_P', 'C_c'), 2)])
        self.assertEqual(reaction.equation, actual_equation)

        self.assertEqual(reaction.properties['lower_flux'], -10)
        self.assertEqual(reaction.properties['upper_flux'], 1000)

    def test_biomass_reaction_exists(self):
        reaction = self.reader.get_reaction('R_Biomass')
        self.assertFalse(reaction.reversible)

        # Compare equation of reaction
        actual_equation = Reaction(Reaction.Right,
                                   [(Compound('M_Glucose_6_P', 'C_c'),
                                     Decimal('0.56'))],
                                   [(Compound('M_Biomass', 'C_c'), 1)])
        self.assertEqual(reaction.equation, actual_equation)

        self.assertEqual(reaction.properties['lower_flux'], 0)
        self.assertEqual(reaction.properties['upper_flux'], 1000)

    def test_objective_exists(self):
        objectives = {entry.id: entry for entry in self.reader.objectives}
        self.assertEqual(len(objectives), 1)

        objective = objectives['obj1']
        self.assertEqual(objective.name, 'Objective 1')
        self.assertEqual(objective.type, 'maximize')
        self.assertEqual(dict(objective.reactions), {'R_Biomass': 1})

    def test_active_objective(self):
        objectives = {entry.id: entry for entry in self.reader.objectives}
        self.assertEqual(self.reader.get_active_objective(),
                         objectives['obj1'])

    def test_flux_bounds_not_present(self):
        flux_bounds = list(self.reader.flux_bounds)
        self.assertEqual(len(flux_bounds), 0)


if __name__ == '__main__':
    unittest.main()



# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2014-2015  Jon Lund Steffensen <jon_steffensen@uri.edu>

import logging

from ..command import Command, SolverCommandMixin, CommandError
from .. import fluxanalysis, fluxcoupling

logger = logging.getLogger(__name__)


class FluxCouplingCommand(SolverCommandMixin, Command):
    """Find flux coupled reactions in the model.

    This identifies any reaction pairs where the flux of one reaction
    constrains the flux of another reaction. The reactions can be coupled in
    three distinct ways depending on the ratio between the reaction fluxes.
    The reactions can be fully coupled (the ratio is static and non-zero);
    partially coupled (the ratio is bounded and non-zero); or directionally
    coupled (the ratio is non-zero).
    """

    def run(self):
        solver = self._get_solver()

        max_reaction = self._model.get_biomass_reaction()
        if max_reaction is None:
            raise CommandError('The biomass reaction was not specified')

        fba_fluxes = dict(fluxanalysis.flux_balance(
            self._mm, max_reaction, tfba=False, solver=solver))
        optimum = fba_fluxes[max_reaction]

        self._fcp = fluxcoupling.FluxCouplingProblem(
            self._mm, {max_reaction: 0.999 * optimum}, solver)

        self._coupled = {}
        self._groups = []

        reactions = sorted(self._mm.reactions)
        for i, reaction1 in enumerate(reactions):
            if reaction1 in self._coupled:
                continue

            for reaction2 in reactions[i+1:]:
                if (reaction2 in self._coupled and
                        (self._coupled[reaction2] ==
                         self._coupled.get(reaction1))):
                    continue

                self._check_reactions(reaction1, reaction2)

        logger.info('Coupled groups:')
        for i, group in enumerate(self._groups):
            if group is not None:
                logger.info('{}: {}'.format(i, ', '.join(sorted(group))))

    def _check_reactions(self, reaction1, reaction2):
        logger.debug('Solving for {}, {}'.format(reaction1, reaction2))
        lower, upper = self._fcp.solve(reaction1, reaction2)

        logger.debug('Result: {}, {}'.format(lower, upper))

        coupling = fluxcoupling.classify_coupling((lower, upper))
        if coupling in (fluxcoupling.CouplingClass.DirectionalForward,
                        fluxcoupling.CouplingClass.DirectionalReverse):
            text = 'Directional, v1 / v2 in [{}, {}]'.format(lower, upper)
            if (coupling == fluxcoupling.CouplingClass.DirectionalReverse and
                    not self._mm.is_reversible(reaction1) and lower == 0.0):
                return
        elif coupling == fluxcoupling.CouplingClass.Full:
            text = 'Full: v1 / v2 = {}'.format(lower)
            self._couple_reactions(reaction1, reaction2)
        elif coupling == fluxcoupling.CouplingClass.Partial:
            text = 'Partial: v1 / v2 in [{}; {}]'.format(lower, upper)
            self._couple_reactions(reaction1, reaction2)
        else:
            return

        print('{}\t{}\t{}\t{}\t{}'.format(
            reaction1, reaction2, lower, upper, text))

    def _couple_reactions(self, reaction1, reaction2):
        logger.debug('Couple {} and {}'.format(reaction1, reaction2))

        if reaction1 in self._coupled and reaction2 in self._coupled:
            if self._coupled[reaction1] == self._coupled[reaction2]:
                return
            logger.debug('Merge groups {}, {}'.format(
                self._coupled[reaction1], self._coupled[reaction2]))
            group_index = len(self._groups)
            group = (self._groups[self._coupled[reaction1]] |
                     self._groups[self._coupled[reaction2]])
            logger.debug('New group is {}: {}'.format(
                group_index, sorted(group)))
            self._groups.append(group)
            self._groups[self._coupled[reaction1]] = None
            self._groups[self._coupled[reaction2]] = None
            for reaction in group:
                self._coupled[reaction] = group_index
        elif reaction1 in self._coupled:
            group_index = self._coupled[reaction1]
            group = self._groups[group_index]
            logger.debug('Put {} into existing group {}: {}'.format(
                reaction2, self._coupled[reaction1], group))
            self._coupled[reaction2] = group_index
            group.add(reaction2)
        elif reaction2 in self._coupled:
            group_index = self._coupled[reaction2]
            group = self._groups[group_index]
            logger.debug('Put {} into existing group {}: {}'.format(
                reaction1, self._coupled[reaction2], group))
            self._coupled[reaction1] = group_index
            group.add(reaction1)
        else:
            group = set([reaction1, reaction2])
            group_index = len(self._groups)
            logger.debug('Creating new group {}'.format(group_index))
            self._coupled[reaction1] = group_index
            self._coupled[reaction2] = group_index
            self._groups.append(group)
